from datetime import datetime
import os
from pathlib import Path
from typing import Any, Tuple
from unittest.mock import Mock

import yaml

import anyscale
from anyscale.client.openapi_client.models.project import Project  # type: ignore
from anyscale.client.openapi_client.models.project_response import ProjectResponse  # type: ignore
from anyscale.client.openapi_client.models.session import Session  # type: ignore
from anyscale.connect import SessionBuilder

UP_CMD = ["anyscale", "up"]


def _make_test_builder(tmp_path: Path) -> Tuple[Any, Any, Any, Any]:
    scratch = tmp_path / "scratch"
    sdk = Mock()
    sess_resp = Mock()
    ray = Mock()
    sess_resp.results = [
        Session(
            id="session_id",
            name="session-1",
            created_at=datetime.now(),
            snapshots_history=[],
            idle_timeout=120,
            tensorboard_available=False,
            project_id="project_id",
            state="Running",
            service_proxy_url="http://session-foo.userdata.com/auth?token=value&bar",
            last_activity_at=datetime.now(),
        )
    ]
    sdk.search_sessions.return_value = sess_resp
    subprocess = Mock()
    builder = SessionBuilder(
        scratch_dir=scratch.absolute().as_posix(),
        anyscale_sdk=sdk,
        subprocess=subprocess,
        ray=ray,
    )
    return builder, sdk, subprocess, ray


def test_new_proj_connect_params(tmp_path: Path, project_test_data: Project) -> None:
    project_dir = (tmp_path / "my_proj").absolute().as_posix()
    builder, sdk, subprocess, ray = _make_test_builder(tmp_path)
    sdk.create_project.return_value = ProjectResponse(result=project_test_data)

    # Should create a new .anyscale.yaml file
    builder.project_dir(project_dir).connect()

    assert anyscale.project.get_project_id(project_dir)
    subprocess.check_call.assert_called_once_with(
        UP_CMD + ["--cloud-name", "anyscale_default_cloud", "session-1"],
        cwd=project_dir,
    )

    # Also check connection params in this test.
    ray.util.connect.assert_called_once_with(
        "session-foo.userdata.com:8081",
        metadata=[("cookie", "anyscale-token=value"), ("port", "10001")],
        secure=False,
    )


def test_detect_existing_proj(tmp_path: Path) -> None:
    nested_dir = (tmp_path / "my_proj" / "nested").absolute().as_posix()
    parent_dir = os.path.dirname(nested_dir)
    os.makedirs(nested_dir)
    builder, sdk, subprocess, ray = _make_test_builder(tmp_path)

    # Setup project in parent dir
    project_yaml = os.path.join(parent_dir, ".anyscale.yaml")
    with open(project_yaml, "w+") as f:
        f.write(yaml.dump({"project_id": 12345}))

    # Should detect the parent project dir
    cwd = os.getcwd()
    try:
        os.chdir(nested_dir)
        builder.connect()
    finally:
        os.chdir(cwd)

    subprocess.check_call.assert_called_once_with(
        UP_CMD + ["--cloud-name", "anyscale_default_cloud", "session-1"],
        cwd=parent_dir,
    )


def test_fallback_scratch_dir(tmp_path: Path, project_test_data: Project) -> None:
    scratch_dir = (tmp_path / "scratch").absolute().as_posix()
    builder, sdk, subprocess, ray = _make_test_builder(tmp_path)
    sdk.create_project.return_value = ProjectResponse(result=project_test_data)

    # Should create a new .anyscale.yaml file in the scratch dir
    builder.connect()

    assert anyscale.project.get_project_id(scratch_dir)
    subprocess.check_call.assert_called_once_with(
        UP_CMD + ["--cloud-name", "anyscale_default_cloud", "session-1"],
        cwd=scratch_dir,
    )



import asyncio
from typing import Any, cast, Dict, Set, Tuple
import urllib

import aioredis
from terminado import NamedTermManager, TermSocket
import tornado.web


class AnyscaleTermManager(NamedTermManager):  # type: ignore
    def __init__(self, **kwargs: Any) -> None:
        super(AnyscaleTermManager, self).__init__(**kwargs)
        # For each terminal, record if it is a debug terminal
        # (keyed by terminal name).
        self.is_debugger: Dict[str, bool] = {}

    def get_terminal(self, term_name: str, is_debugger: bool = False) -> Any:
        terminal = super(AnyscaleTermManager, self).get_terminal(term_name)
        self.is_debugger[term_name] = is_debugger
        return terminal

    def new_named_terminal(self, **kwargs: Any) -> Tuple[str, Any]:
        name, term = super(AnyscaleTermManager, self).new_named_terminal(**kwargs)
        self.is_debugger[name] = False
        return name, term


class AnyscaleTermSocket(TermSocket):  # type: ignore
    def check_origin(self, origin: Any) -> bool:
        parsed_origin = urllib.parse.urlparse(origin)
        if self.application.settings["deploy_environment"] == "development":
            return True
        elif self.application.settings["deploy_environment"] == "staging":
            return cast(bool, parsed_origin.netloc.endswith("anyscale.dev"))
        else:
            return cast(bool, parsed_origin.netloc.endswith("beta.anyscale.com"))


class TerminalPageHandler(tornado.web.RequestHandler):
    """Render the /ttyX pages"""

    def delete(self, term_name: str) -> None:
        term_manager = self.application.settings["term_manager"]

        term_manager.kill(term_name)
        del term_manager.terminals[term_name]


class TerminalListHandler(tornado.web.RequestHandler):
    """List active terminals."""

    def get(self) -> None:
        term_manager = self.application.settings["term_manager"]

        data = {
            "terminals": [
                {"id": term_name, "debugger": term_manager.is_debugger[term_name]}
                for term_name in term_manager.terminals.keys()
            ]
        }

        self.write(data)


class NewTerminalHandler(tornado.web.RequestHandler):
    """Create new unused terminal"""

    def get(self) -> None:
        self.application.settings["term_manager"].new_named_terminal()


def make_application(deploy_environment: str) -> tornado.web.Application:
    term_manager = AnyscaleTermManager(shell_command=["bash"], max_terminals=100)

    handlers: Any = [
        (
            r"/webterminal/_websocket/(\w+)",
            AnyscaleTermSocket,
            {"term_manager": term_manager},
        ),
        (r"/webterminal/new/?", NewTerminalHandler),
        (r"/webterminal/list", TerminalListHandler),
        (r"/webterminal/(\w+)/?", TerminalPageHandler),
    ]
    return tornado.web.Application(
        handlers, term_manager=term_manager, deploy_environment=deploy_environment,
    )


async def initialize_debugger(application: Any) -> None:
    redis_client = await aioredis.create_redis_pool(
        ("localhost", 6379), password="5241590000000000", encoding="utf-8"
    )

    ray_version = await redis_client.execute("get", b"VERSION_INFO")
    print("Starting debugger backend, Ray version is {}".format(ray_version))

    registered_breakpoints: Set[str] = set()

    while True:
        keys = await redis_client.keys(b"RAY_PDB_*")

        new_breakpoints = set(keys) - registered_breakpoints

        if len(new_breakpoints) > 0:
            print("new breakpoints", new_breakpoints)

        for new_breakpoint in new_breakpoints:
            term_manager = application.settings["term_manager"]
            debugger = term_manager.get_terminal(new_breakpoint, is_debugger=True)
            debugger.ptyproc.write("ray debug\n")

        registered_breakpoints = registered_breakpoints.union(new_breakpoints)

        await asyncio.sleep(1.0)


def main(deploy_environment: str, use_debugger: bool) -> None:
    application = make_application(deploy_environment)

    port = 8700
    application.listen(port, "localhost")
    print("Listening on localhost:{}".format(port))
    loop = tornado.ioloop.IOLoop.instance()

    if use_debugger:
        loop.run_sync(lambda: initialize_debugger(application))

    try:
        loop.start()
    except KeyboardInterrupt:
        print(" Shutting down on SIGINT")
    finally:
        application.settings["term_manager"].shutdown()
        loop.close()



import anyscale.sdk.anyscale_client as anyscale_client
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi


class AnyscaleSDK(DefaultApi):  # type: ignore
    def __init__(self, auth_token: str, host: str = "https://beta.anyscale.com/ext"):
        configuration = anyscale_client.Configuration(host=host)
        api_client = anyscale_client.ApiClient(
            configuration, cookie=f"cli_token={auth_token}"
        )

        super(AnyscaleSDK, self).__init__(api_client)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SessionFinishUpOptions(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'startup_log': 'str',
        'new_session': 'bool',
        'head_node_ip': 'str',
        'cluster_launch_errored': 'bool'
    }

    attribute_map = {
        'startup_log': 'startup_log',
        'new_session': 'new_session',
        'head_node_ip': 'head_node_ip',
        'cluster_launch_errored': 'cluster_launch_errored'
    }

    def __init__(self, startup_log=None, new_session=None, head_node_ip=None, cluster_launch_errored=None, local_vars_configuration=None):  # noqa: E501
        """SessionFinishUpOptions - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._startup_log = None
        self._new_session = None
        self._head_node_ip = None
        self._cluster_launch_errored = None
        self.discriminator = None

        if startup_log is not None:
            self.startup_log = startup_log
        self.new_session = new_session
        self.head_node_ip = head_node_ip
        self.cluster_launch_errored = cluster_launch_errored

    @property
    def startup_log(self):
        """Gets the startup_log of this SessionFinishUpOptions.  # noqa: E501


        :return: The startup_log of this SessionFinishUpOptions.  # noqa: E501
        :rtype: str
        """
        return self._startup_log

    @startup_log.setter
    def startup_log(self, startup_log):
        """Sets the startup_log of this SessionFinishUpOptions.


        :param startup_log: The startup_log of this SessionFinishUpOptions.  # noqa: E501
        :type: str
        """

        self._startup_log = startup_log

    @property
    def new_session(self):
        """Gets the new_session of this SessionFinishUpOptions.  # noqa: E501


        :return: The new_session of this SessionFinishUpOptions.  # noqa: E501
        :rtype: bool
        """
        return self._new_session

    @new_session.setter
    def new_session(self, new_session):
        """Sets the new_session of this SessionFinishUpOptions.


        :param new_session: The new_session of this SessionFinishUpOptions.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and new_session is None:  # noqa: E501
            raise ValueError("Invalid value for `new_session`, must not be `None`")  # noqa: E501

        self._new_session = new_session

    @property
    def head_node_ip(self):
        """Gets the head_node_ip of this SessionFinishUpOptions.  # noqa: E501


        :return: The head_node_ip of this SessionFinishUpOptions.  # noqa: E501
        :rtype: str
        """
        return self._head_node_ip

    @head_node_ip.setter
    def head_node_ip(self, head_node_ip):
        """Sets the head_node_ip of this SessionFinishUpOptions.


        :param head_node_ip: The head_node_ip of this SessionFinishUpOptions.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and head_node_ip is None:  # noqa: E501
            raise ValueError("Invalid value for `head_node_ip`, must not be `None`")  # noqa: E501

        self._head_node_ip = head_node_ip

    @property
    def cluster_launch_errored(self):
        """Gets the cluster_launch_errored of this SessionFinishUpOptions.  # noqa: E501


        :return: The cluster_launch_errored of this SessionFinishUpOptions.  # noqa: E501
        :rtype: bool
        """
        return self._cluster_launch_errored

    @cluster_launch_errored.setter
    def cluster_launch_errored(self, cluster_launch_errored):
        """Sets the cluster_launch_errored of this SessionFinishUpOptions.


        :param cluster_launch_errored: The cluster_launch_errored of this SessionFinishUpOptions.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and cluster_launch_errored is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_launch_errored`, must not be `None`")  # noqa: E501

        self._cluster_launch_errored = cluster_launch_errored

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionFinishUpOptions):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionFinishUpOptions):
            return True

        return self.to_dict() != other.to_dict()



import json
import os
import shutil
import sys
from typing import Any, Dict, Optional, Tuple

import click
from click import ClickException
import jsonschema
from openapi_client.rest import ApiException  # type: ignore
import yaml

import anyscale
from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
from anyscale.util import (
    format_api_exception,
    get_endpoint,
    send_json_request,
    slugify,
    validate_cluster_configuration,
)

ANYSCALE_PROJECT_FILE = ".anyscale.yaml"
ANYSCALE_AUTOSCALER_FILE = "session-default.yaml"


CLUSTER_YAML_TEMPLATE = """
# This file was generated by `anyscale init`.

cluster_name: default

max_workers: 1

docker:
    image: anyscale/ray-ml:latest
    container_name: ray_container
    pull_before_run: true

setup_commands: []

head_setup_commands: []

# Cloud-provider specific configuration.
provider:
    type: aws
    region: us-west-2
    availability_zone: us-west-2a

# How Ray will authenticate with newly launched nodes.
auth:
    ssh_user: ubuntu

rsync_exclude:
    - "**/.git"
    - "**/.git/**"

rsync_filter:
    - ".gitignore"
"""

CLUSTER_CONFIG_TEMPLATE_STR = json.dumps(
    yaml.load(CLUSTER_YAML_TEMPLATE, Loader=yaml.SafeLoader)
)


def validate_project_schema(project_config: Dict[str, str]) -> Any:
    """Validate a project config against the project schema.
    Args:
        project_config (dict): Parsed project yaml.
    Raises:
        jsonschema.exceptions.ValidationError: This exception is raised
            if the project file is not valid.
    """
    dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(dir, "ProjectConfig.json")) as f:
        schema = json.load(f)

    jsonschema.validate(instance=project_config, schema=schema)


def find_project_root(directory: str) -> Optional[str]:
    """Find root directory of the project.

    Args:
        directory (str): Directory to start the search in.

    Returns:
        Path of the parent directory containing the project
        or None if no such project is found.
    """
    prev, directory = None, os.path.abspath(directory)
    while prev != directory:
        if os.path.exists(os.path.join(directory, ANYSCALE_PROJECT_FILE)):
            return directory
        prev, directory = directory, os.path.abspath(os.path.join(directory, os.pardir))
    return None


class ProjectDefinition(object):
    def __init__(self, root_dir: str):
        self.root = os.path.join(root_dir, "")
        anyscale_yaml = os.path.join(root_dir, ANYSCALE_PROJECT_FILE)
        if os.path.exists(anyscale_yaml):
            with open(anyscale_yaml) as f:
                self.config = yaml.safe_load(f)
        else:
            self.config = {}

        if "cluster" not in self.config:
            self.config["cluster"] = {
                "config": os.path.join(self.root, ANYSCALE_AUTOSCALER_FILE)
            }

    def cluster_yaml(self) -> str:
        if "config" in self.config["cluster"]:
            return str(self.config["cluster"]["config"])
        return os.path.join(self.root, ANYSCALE_AUTOSCALER_FILE)


def load_project_or_throw() -> ProjectDefinition:
    # First check if there is a .anyscale.yaml.
    root_dir = find_project_root(os.getcwd())
    if not root_dir:
        raise ClickException("No project directory found")
    return ProjectDefinition(root_dir)


def get_project_id(project_dir: str) -> str:
    """
    Args:
        project_dir: Project root directory.

    Returns:
        The ID of the associated Project in the database.

    Raises:
        ValueError: If the current project directory does
            not contain a project ID.
    """
    project_filename = os.path.join(project_dir, ANYSCALE_PROJECT_FILE)
    if os.path.isfile(project_filename):
        with open(project_filename) as f:
            config = yaml.safe_load(f)
            project_id = config["project_id"]
    else:
        # TODO(pcm): Consider doing this for the user and retrying the command
        # they were trying to run.
        raise ClickException(
            "Ray project in {} not registered yet. "
            "Did you run 'anyscale init'?".format(project_dir)
        )
    try:
        result = str(project_id)
    except ValueError:
        # TODO(pcm): Tell the user what to do here.
        raise ClickException(
            "{} does not contain a valid project ID".format(project_filename)
        )
    return result


def validate_project_name(project_name: str) -> bool:
    return " " not in project_name.strip()


def get_project_sessions(
    project_id: str, session_name: Optional[str], api_client: DefaultApi = None
) -> Any:
    if api_client is None:
        return _get_project_sessions(project_id, session_name)

    with format_api_exception(ApiException):
        response = api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, name_match=session_name, active_only=True
        )
    sessions = response.results
    if len(sessions) == 0:
        raise ClickException(
            "No active session matching pattern {} found".format(session_name)
        )
    return sessions


# TODO (jbai): DEPRECATED - will be removed when OpenApi migration is completed
def _get_project_sessions(project_id: str, session_name: Optional[str]) -> Any:
    response = send_json_request(
        "/api/v2/sessions/",
        {"project_id": project_id, "name_match": session_name, "active_only": True},
    )
    sessions = response["results"]
    if len(sessions) == 0:
        raise ClickException(
            "No active session matching pattern {} found".format(session_name)
        )
    return sessions


def get_project_session(
    project_id: str, session_name: Optional[str], api_client: DefaultApi = None
) -> Any:
    if api_client is None:
        return _get_project_session(project_id, session_name)

    sessions = get_project_sessions(project_id, session_name, api_client)
    if len(sessions) > 1:
        raise ClickException(
            "Multiple active sessions: {}\n"
            "Please specify the one you want to refer to.".format(
                [session.name for session in sessions]
            )
        )
    return sessions[0]


# TODO (jbai): DEPRECATED - will be removed when OpenApi migration is completed
def _get_project_session(project_id: str, session_name: Optional[str]) -> Any:
    sessions = get_project_sessions(project_id, session_name)
    if len(sessions) > 1:
        raise ClickException(
            "Multiple active sessions: {}\n"
            "Please specify the one you want to refer to.".format(
                [session["name"] for session in sessions]
            )
        )
    return sessions[0]


def get_proj_name_from_id(project_id: str, api_client: DefaultApi) -> str:
    with format_api_exception(ApiException):
        resp = api_client.get_project_api_v2_projects_project_id_get(
            project_id=project_id
        )

    if resp is None:
        raise ClickException(
            "This local project is not registered with anyscale. Please re-run `anyscale init`."
        )
    else:
        return str(resp.result.name)


def get_proj_id_from_name(project_name: str, api_client: DefaultApi) -> str:
    with format_api_exception(ApiException):
        resp = api_client.find_project_by_project_name_api_v2_projects_find_by_name_get(
            name=project_name
        )

    if not resp.results:
        raise ClickException(
            f"There is no project '{project_name}' that is registered with Anyscale. "
            "View the registered projects with `anyscale list projects`."
        )

    project = resp.results[0]
    return str(project.id)


def clone_cluster_config(
    project_name: str, directory: str, project_id: str, api_client: DefaultApi
) -> None:
    """
    Copy the cluster config from the latest session if available. Otherwise copy the
    cluster config from the project.
    """
    with format_api_exception(ApiException):
        sessions = api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id
        ).results

    if len(sessions) > 0:
        lastest_session = sessions[0]

        with format_api_exception(ApiException):
            cluster_config = api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get(
                lastest_session.id
            ).result.config
    else:
        with format_api_exception(ApiException):
            cluster_config = api_client.get_project_latest_cluster_config_api_v2_projects_project_id_latest_cluster_config_get(
                project_id
            ).result.config

    _write_cluster_config_to_disk(project_id, cluster_config, directory)


def _write_cluster_config_to_disk(
    project_id: str, cluster_config: str, directory: str
) -> None:
    with open("{}/{}".format(directory, ANYSCALE_PROJECT_FILE), "w") as f:
        f.write("{}".format("project_id: {}".format(project_id)))

    cluster_config_json = json.loads(cluster_config)
    with open("{}/{}".format(directory, ANYSCALE_AUTOSCALER_FILE), "w") as f:
        yaml.dump(cluster_config_json, f, default_flow_style=False)


def create_new_proj_def(
    name: Optional[str],
    cluster_config_file: Optional[str],
    use_default_yaml: bool = True,
    api_client: DefaultApi = None,
) -> Tuple[str, ProjectDefinition]:
    project_name = ""
    if not name:
        while project_name == "":
            project_name = click.prompt("Project name", type=str)
            if not validate_project_name(project_name):
                print(
                    '"{}" contains spaces. Please enter a project name without spaces'.format(
                        project_name
                    ),
                    file=sys.stderr,
                )
                project_name = ""
        if not cluster_config_file:
            # TODO (yiran): Print cluster.yaml path in the else case.
            cluster_config_file = click.prompt(
                "Cluster yaml file (optional)",
                type=click.Path(exists=True),
                default=".",
                show_default=False,
            )
            if cluster_config_file == ".":
                # handling default value from prompt
                cluster_config_file = None
            else:
                use_default_yaml = False
    else:
        project_name = str(name)
    if slugify(project_name) != project_name:
        project_name = slugify(project_name)
        print("Normalized project name to {}".format(project_name))

    # Create startup.yaml.
    if cluster_config_file:
        validate_cluster_configuration(cluster_config_file, api_instance=api_client)
        if use_default_yaml and (
            not os.path.exists(anyscale.project.ANYSCALE_AUTOSCALER_FILE)
            or not os.path.samefile(
                cluster_config_file, anyscale.project.ANYSCALE_AUTOSCALER_FILE
            )
        ):
            shutil.copyfile(
                cluster_config_file, anyscale.project.ANYSCALE_AUTOSCALER_FILE
            )
    else:
        if not os.path.exists(anyscale.project.ANYSCALE_AUTOSCALER_FILE):
            with open(anyscale.project.ANYSCALE_AUTOSCALER_FILE, "w") as f:
                f.write(anyscale.project.CLUSTER_YAML_TEMPLATE)
    project_definition = anyscale.project.ProjectDefinition(os.getcwd())
    project_definition.config["name"] = project_name
    if not use_default_yaml and cluster_config_file:
        project_definition.config["cluster"]["config"] = os.path.join(
            os.getcwd(), str(cluster_config_file)
        )
    return project_name, project_definition


def register_project(
    project_definition: ProjectDefinition, api_client: DefaultApi
) -> None:
    validate_cluster_configuration(
        project_definition.cluster_yaml(), api_instance=api_client
    )

    project_name = project_definition.config["name"]
    description = project_definition.config.get("description", "")

    with open(project_definition.cluster_yaml(), "r") as f:
        initial_cluster_config = yaml.load(f, Loader=yaml.SafeLoader)

    # Add a database entry for the new Project.
    with format_api_exception(ApiException):
        resp = api_client.create_project_api_v2_projects_post(
            write_project={
                "name": project_name,
                "description": description,
                "initial_cluster_config": json.dumps(initial_cluster_config),
            }
        )
    result = resp.result
    project_id = result.id

    with open(anyscale.project.ANYSCALE_PROJECT_FILE, "w") as f:
        yaml.dump(
            {
                "project_id": project_id,
                "cluster": {"config": project_definition.cluster_yaml()},
            },
            f,
        )

    # Print success message
    url = get_endpoint(f"/projects/{project_id}")
    print(f"Project {project_id} created. View at {url}")



# coding: utf-8

# flake8: noqa
"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from openapi_client.models.aws_credentials import AWSCredentials
from openapi_client.models.add_instance_pool_member import AddInstancePoolMember
from openapi_client.models.anyscale_aws_account import AnyscaleAWSAccount
from openapi_client.models.anyscale_version_response import AnyscaleVersionResponse
from openapi_client.models.anyscaleawsaccount_response import AnyscaleawsaccountResponse
from openapi_client.models.anyscaleversionresponse_response import AnyscaleversionresponseResponse
from openapi_client.models.applied_snapshot import AppliedSnapshot
from openapi_client.models.archived_logs_info import ArchivedLogsInfo
from openapi_client.models.archivedlogsinfo_response import ArchivedlogsinfoResponse
from openapi_client.models.autoscaler_credentials import AutoscalerCredentials
from openapi_client.models.autoscalercredentials_response import AutoscalercredentialsResponse
from openapi_client.models.autosync_session_id import AutosyncSessionId
from openapi_client.models.autosyncsessionid_list_response import AutosyncsessionidListResponse
from openapi_client.models.autosyncsessionid_response import AutosyncsessionidResponse
from openapi_client.models.awscredentials_response import AwscredentialsResponse
from openapi_client.models.batch_response_batched_result_organization_invitation_base import BatchResponseBatchedResultOrganizationInvitationBase
from openapi_client.models.batched_result_organization_invitation_base import BatchedResultOrganizationInvitationBase
from openapi_client.models.change_password_params import ChangePasswordParams
from openapi_client.models.cloud import Cloud
from openapi_client.models.cloud_config import CloudConfig
from openapi_client.models.cloud_list_response import CloudListResponse
from openapi_client.models.cloud_name_options import CloudNameOptions
from openapi_client.models.cloud_providers import CloudProviders
from openapi_client.models.cloud_response import CloudResponse
from openapi_client.models.cloud_types import CloudTypes
from openapi_client.models.cluster_config import ClusterConfig
from openapi_client.models.cluster_status import ClusterStatus
from openapi_client.models.clusterconfig_response import ClusterconfigResponse
from openapi_client.models.clusterstatus_response import ClusterstatusResponse
from openapi_client.models.create_from_github_options import CreateFromGithubOptions
from openapi_client.models.create_nodes_options import CreateNodesOptions
from openapi_client.models.create_organization_invitation import CreateOrganizationInvitation
from openapi_client.models.create_session_from_snapshot_options import CreateSessionFromSnapshotOptions
from openapi_client.models.create_snapshot_options import CreateSnapshotOptions
from openapi_client.models.create_user import CreateUser
from openapi_client.models.create_user_project_collaborator import CreateUserProjectCollaborator
from openapi_client.models.create_user_project_collaborator_value import CreateUserProjectCollaboratorValue
from openapi_client.models.error import Error
from openapi_client.models.execute_command_response import ExecuteCommandResponse
from openapi_client.models.execute_interactive_command_options import ExecuteInteractiveCommandOptions
from openapi_client.models.execute_shell_command_options import ExecuteShellCommandOptions
from openapi_client.models.executecommandresponse_response import ExecutecommandresponseResponse
from openapi_client.models.external_service_status import ExternalServiceStatus
from openapi_client.models.external_service_status_response import ExternalServiceStatusResponse
from openapi_client.models.externalservicestatusresponse_response import ExternalservicestatusresponseResponse
from openapi_client.models.feature_flag_response import FeatureFlagResponse
from openapi_client.models.featureflagresponse_response import FeatureflagresponseResponse
from openapi_client.models.gateway_request import GatewayRequest
from openapi_client.models.gatewayrequest_response import GatewayrequestResponse
from openapi_client.models.github_project import GithubProject
from openapi_client.models.githubproject_response import GithubprojectResponse
from openapi_client.models.http_validation_error import HTTPValidationError
from openapi_client.models.head_ip import HeadIp
from openapi_client.models.headip_response import HeadipResponse
from openapi_client.models.instance import Instance
from openapi_client.models.instance_external_ip import InstanceExternalIP
from openapi_client.models.instance_id import InstanceID
from openapi_client.models.instance_internal_ip import InstanceInternalIP
from openapi_client.models.instance_is_running import InstanceIsRunning
from openapi_client.models.instance_is_terminated import InstanceIsTerminated
from openapi_client.models.instance_list_response import InstanceListResponse
from openapi_client.models.instance_pool_member import InstancePoolMember
from openapi_client.models.instance_response import InstanceResponse
from openapi_client.models.instance_status import InstanceStatus
from openapi_client.models.instanceexternalip_response import InstanceexternalipResponse
from openapi_client.models.instanceid_response import InstanceidResponse
from openapi_client.models.instanceinternalip_response import InstanceinternalipResponse
from openapi_client.models.instanceisrunning_response import InstanceisrunningResponse
from openapi_client.models.instanceisterminated_response import InstanceisterminatedResponse
from openapi_client.models.instancepoolmember_response import InstancepoolmemberResponse
from openapi_client.models.json_dumped_gateway_response_message import JsonDumpedGatewayResponseMessage
from openapi_client.models.json_patch_operation import JsonPatchOperation
from openapi_client.models.list_response_metadata import ListResponseMetadata
from openapi_client.models.login_user_params import LoginUserParams
from openapi_client.models.logs_output import LogsOutput
from openapi_client.models.logsoutput_response import LogsoutputResponse
from openapi_client.models.nodes_options import NodesOptions
from openapi_client.models.non_terminated_nodes_options import NonTerminatedNodesOptions
from openapi_client.models.organization import Organization
from openapi_client.models.organization_availability import OrganizationAvailability
from openapi_client.models.organization_collaborator import OrganizationCollaborator
from openapi_client.models.organization_invitation import OrganizationInvitation
from openapi_client.models.organization_invitation_base import OrganizationInvitationBase
from openapi_client.models.organization_permission_level import OrganizationPermissionLevel
from openapi_client.models.organization_project_collaborator import OrganizationProjectCollaborator
from openapi_client.models.organization_project_collaborator_value import OrganizationProjectCollaboratorValue
from openapi_client.models.organization_response import OrganizationResponse
from openapi_client.models.organizationavailability_response import OrganizationavailabilityResponse
from openapi_client.models.organizationcollaborator_list_response import OrganizationcollaboratorListResponse
from openapi_client.models.organizationinvitation_list_response import OrganizationinvitationListResponse
from openapi_client.models.organizationinvitation_response import OrganizationinvitationResponse
from openapi_client.models.organizationinvitationbase_response import OrganizationinvitationbaseResponse
from openapi_client.models.organizationprojectcollaborator_list_response import OrganizationprojectcollaboratorListResponse
from openapi_client.models.permission_level import PermissionLevel
from openapi_client.models.project import Project
from openapi_client.models.project_base import ProjectBase
from openapi_client.models.project_collaborator import ProjectCollaborator
from openapi_client.models.project_collaborator_value import ProjectCollaboratorValue
from openapi_client.models.project_create_message import ProjectCreateMessage
from openapi_client.models.project_default_session_name import ProjectDefaultSessionName
from openapi_client.models.project_delete_message import ProjectDeleteMessage
from openapi_client.models.project_list_response import ProjectListResponse
from openapi_client.models.project_patch_message import ProjectPatchMessage
from openapi_client.models.project_response import ProjectResponse
from openapi_client.models.projectbase_response import ProjectbaseResponse
from openapi_client.models.projectcollaborator_list_response import ProjectcollaboratorListResponse
from openapi_client.models.projectdefaultsessionname_response import ProjectdefaultsessionnameResponse
from openapi_client.models.query_pool_size import QueryPoolSize
from openapi_client.models.request_instance_pool_member import RequestInstancePoolMember
from openapi_client.models.request_organization_identifiers_params import RequestOrganizationIdentifiersParams
from openapi_client.models.request_password_reset_params import RequestPasswordResetParams
from openapi_client.models.reset_password_params import ResetPasswordParams
from openapi_client.models.s3_download_location import S3DownloadLocation
from openapi_client.models.server_session_token import ServerSessionToken
from openapi_client.models.serversessiontoken_response import ServersessiontokenResponse
from openapi_client.models.session import Session
from openapi_client.models.session_autosync_sessions_update_message import SessionAutosyncSessionsUpdateMessage
from openapi_client.models.session_command import SessionCommand
from openapi_client.models.session_command_finish_options import SessionCommandFinishOptions
from openapi_client.models.session_command_id import SessionCommandId
from openapi_client.models.session_describe import SessionDescribe
from openapi_client.models.session_details import SessionDetails
from openapi_client.models.session_execute_message import SessionExecuteMessage
from openapi_client.models.session_finish_command_message import SessionFinishCommandMessage
from openapi_client.models.session_finish_up_options import SessionFinishUpOptions
from openapi_client.models.session_history_item import SessionHistoryItem
from openapi_client.models.session_id import SessionId
from openapi_client.models.session_kill_command_message import SessionKillCommandMessage
from openapi_client.models.session_list_response import SessionListResponse
from openapi_client.models.session_response import SessionResponse
from openapi_client.models.session_ssh_key import SessionSshKey
from openapi_client.models.session_starting_up_data import SessionStartingUpData
from openapi_client.models.session_state import SessionState
from openapi_client.models.session_state_data import SessionStateData
from openapi_client.models.session_stop_message import SessionStopMessage
from openapi_client.models.session_stopping_data import SessionStoppingData
from openapi_client.models.session_up_options import SessionUpOptions
from openapi_client.models.session_up_response import SessionUpResponse
from openapi_client.models.sessioncommand_list_response import SessioncommandListResponse
from openapi_client.models.sessioncommandid_response import SessioncommandidResponse
from openapi_client.models.sessiondescribe_response import SessiondescribeResponse
from openapi_client.models.sessiondetails_response import SessiondetailsResponse
from openapi_client.models.sessionhistoryitem_list_response import SessionhistoryitemListResponse
from openapi_client.models.sessionid_response import SessionidResponse
from openapi_client.models.sessionsshkey_response import SessionsshkeyResponse
from openapi_client.models.sessionupresponse_response import SessionupresponseResponse
from openapi_client.models.set_node_tags_options import SetNodeTagsOptions
from openapi_client.models.setup_initialize_session_options import SetupInitializeSessionOptions
from openapi_client.models.snapshot import Snapshot
from openapi_client.models.snapshot_create import SnapshotCreate
from openapi_client.models.snapshot_create_message import SnapshotCreateMessage
from openapi_client.models.snapshot_delete import SnapshotDelete
from openapi_client.models.snapshot_delete_message import SnapshotDeleteMessage
from openapi_client.models.snapshot_files import SnapshotFiles
from openapi_client.models.snapshot_id import SnapshotID
from openapi_client.models.snapshot_list import SnapshotList
from openapi_client.models.snapshot_list_item import SnapshotListItem
from openapi_client.models.snapshot_patch_message import SnapshotPatchMessage
from openapi_client.models.snapshot_project_file import SnapshotProjectFile
from openapi_client.models.snapshot_response import SnapshotResponse
from openapi_client.models.snapshotcreate_response import SnapshotcreateResponse
from openapi_client.models.snapshotdelete_response import SnapshotdeleteResponse
from openapi_client.models.snapshotfiles_response import SnapshotfilesResponse
from openapi_client.models.snapshotid_response import SnapshotidResponse
from openapi_client.models.snapshotlist_response import SnapshotlistResponse
from openapi_client.models.socket_message_schemas import SocketMessageSchemas
from openapi_client.models.socket_message_types import SocketMessageTypes
from openapi_client.models.start_empty_session_response import StartEmptySessionResponse
from openapi_client.models.start_session_options import StartSessionOptions
from openapi_client.models.startemptysessionresponse_response import StartemptysessionresponseResponse
from openapi_client.models.stop_session_options import StopSessionOptions
from openapi_client.models.take_snapshot_options import TakeSnapshotOptions
from openapi_client.models.update_organization_collaborator import UpdateOrganizationCollaborator
from openapi_client.models.update_project_collaborator import UpdateProjectCollaborator
from openapi_client.models.upload_session_command_logs_locations import UploadSessionCommandLogsLocations
from openapi_client.models.uploadsessioncommandlogslocations_response import UploadsessioncommandlogslocationsResponse
from openapi_client.models.user_info import UserInfo
from openapi_client.models.user_resend_email_options import UserResendEmailOptions
from openapi_client.models.userinfo_response import UserinfoResponse
from openapi_client.models.validation_error import ValidationError
from openapi_client.models.web_terminal import WebTerminal
from openapi_client.models.webterminal_list_response import WebterminalListResponse
from openapi_client.models.write_cloud import WriteCloud
from openapi_client.models.write_cluster_config import WriteClusterConfig
from openapi_client.models.write_project import WriteProject
from openapi_client.models.write_session import WriteSession
from openapi_client.models.write_snapshot import WriteSnapshot



import copy
import json
import os
import subprocess
import sys
import tempfile
from typing import Any, Dict, Optional, Tuple, Union

import click
from openapi_client.rest import ApiException  # type: ignore
import yaml

from anyscale import ANYSCALE_ENV
from anyscale.api import get_api_client
from anyscale.autosync_heartbeat import (
    get_working_dir_host_mount_location,
    managed_autosync_session,
    perform_autopush_synchronization,
    perform_sync,
)
from anyscale.client.openapi_client import CreateSessionFromSnapshotOptions, Session, SessionFinishUpOptions, SessionState, SessionUpOptions, SetupInitializeSessionOptions, StopSessionOptions  # type: ignore
from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
from anyscale.cloud import get_cloud_id_and_name
from anyscale.cluster_config import (
    _configure_min_max_workers,
    configure_for_session,
    get_cluster_config,
)
from anyscale.project import (
    get_project_id,
    get_project_session,
    get_project_sessions,
    load_project_or_throw,
)
from anyscale.util import (
    canonicalize_remote_location,
    download_anyscale_wheel,
    format_api_exception,
    get_container_name,
    get_endpoint,
    get_working_dir,
    install_anyscale_hooks,
    populate_session_args,
    slugify,
    validate_cluster_configuration,
    wait_for_session_start,
)
from anyscale.utils.aws_credentials_util import (
    get_credentials_as_env_vars_from_cluster_config,
)
from anyscale.utils.env_utils import set_env


def get_head_node_ip(cluster_config: Union[Dict[str, Any], str]) -> Any:
    from ray.autoscaler.sdk import get_head_node_ip

    return get_head_node_ip(cluster_config)


def rsync(*args: Any, **kwargs: Any) -> None:
    from ray.autoscaler.sdk import rsync as ray_rsync

    ray_rsync(*args, **kwargs)


class SessionController:
    def __init__(self, api_client: Optional[DefaultApi] = None):
        if api_client is None:
            api_client = get_api_client()
        self.api_client = api_client

    def stop(
        self,
        session_name: Optional[str],
        terminate: bool,
        delete: bool,
        workers_only: bool,
        keep_min_workers: bool,
        skip_snapshot: bool = False,
    ) -> None:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        sessions = get_project_sessions(project_id, session_name, self.api_client)
        terminate = terminate or delete

        if not session_name and len(sessions) > 1:
            raise click.ClickException(
                "Multiple active sessions: {}\n"
                "Please specify the one you want to stop with --session-name.".format(
                    [session["name"] for session in sessions]
                )
            )

        for session in sessions:
            # Stop the session and mark it as stopped in the database.
            self.api_client.stop_session_api_v2_sessions_session_id_stop_post(
                session.id,
                StopSessionOptions(
                    terminate=terminate,
                    workers_only=workers_only,
                    keep_min_workers=keep_min_workers,
                    delete=delete,
                    take_snapshot=not skip_snapshot,
                ),
            )

        session_names = [session.name for session in sessions]
        session_names_str = ", ".join(session_names)
        url = get_endpoint(f"/projects/{project_id}")
        print(f"Session {session_names_str} stopping. View progress at {url}")

    def _load_cluster_config(
        self, config: Optional[str], session_exists: bool, project_definition: Any
    ) -> Optional[Dict[str, Any]]:
        """config option to cluster config.

        Excision of existing logic for C901.
        """
        cluster_config = None
        if not session_exists:
            config_attribute_provided = True
            if not config:
                config_attribute_provided = False
                config = project_definition.cluster_yaml()

            config = str(config)  # for typing.
            if not os.path.exists(config):
                if config_attribute_provided:
                    raise ValueError("Config file {} not found".format(config))
                else:
                    raise ValueError(
                        "No config param provided and default config file session-default.yaml not found. Please provide a config file using the --config attribute."
                    )

        if config:
            with open(config) as f:
                cluster_config_filled = populate_session_args(f.read(), config)
                cluster_config = yaml.safe_load(cluster_config_filled)

            validate_cluster_configuration(config, cluster_config, self.api_client)

        return cluster_config

    def up(
        self,
        session_name: Optional[str],
        config: Optional[str],
        build_id: Optional[str],
        compute_template_id: Optional[str],
        min_workers: Optional[int],
        max_workers: Optional[int],
        no_restart: bool,
        restart_only: bool,
        disable_sync: bool,
        cloud_id: Optional[str],
        cloud_name: Optional[str],
        idle_timeout: Optional[int],
        yes: bool,
        verbose: bool = False,
        no_rapid_start: bool = False,
    ) -> None:
        import ray.scripts.scripts as autoscaler_scripts

        # TODO(ilr) Split this into ~3 functions, look at jbai's comments on #1469

        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)

        if not session_name:
            session_name = str(
                self.api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get(
                    project_id=project_id,
                ).result.name
            )

        session_list = self.api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, active_only=False, name=session_name
        ).results

        session_exists = len(session_list) > 0

        # Deduce whether we are launching this session
        # from a cluster config yaml
        # as opposed to a build ID + compute template ID combination.
        # The logic is:
        #  - If either build or compute template is specified,
        #    we will use those. Both must be specified.
        #  - Otherwise, we will use a cluster config yaml.
        #     - If a yaml name is explicitly specified, we will use that.
        #     - Otherwise, look for session-default.yaml.
        using_cluster_config = build_id is None and compute_template_id is None
        if not using_cluster_config:
            if config:
                raise ValueError(
                    "If launching from a build/compute template, a config file cannot be specified."
                )
            if not build_id or not compute_template_id:
                raise ValueError(
                    "Both build ID and Compute Template must be specified if launching a session that way."
                )

        # If using a user cluster config, load it.
        maybe_cluster_config = None
        if using_cluster_config:
            maybe_cluster_config = self._load_cluster_config(
                config=config,
                session_exists=session_exists,
                project_definition=project_definition,
            )

        if no_rapid_start:
            if not maybe_cluster_config:
                raise ValueError(
                    "Please specify a cluster config if using no_rapid_start."
                )
            maybe_cluster_config["provider"]["bypass_rapid_start"] = True

        cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_id, cloud_name)
        assert cloud_id is not None, "Failed to get cloud."

        with format_api_exception(ApiException):
            up_response = self.api_client.session_up_api_v2_sessions_up_post(
                session_up_options=SessionUpOptions(
                    project_id=project_id,
                    name=session_name,
                    cluster_config={"config": json.dumps(maybe_cluster_config)}
                    if maybe_cluster_config
                    else None,
                    build_id=build_id,
                    compute_template_id=compute_template_id,
                    cloud_id=cloud_id,
                    idle_timeout=idle_timeout,
                )
            ).result
        session_id = up_response.session_id

        # Download & write anyscale wheel!
        download_anyscale_wheel(self.api_client, session_id)

        # Get previous head IP
        with format_api_exception(ApiException):
            ip_resp = self.api_client.get_session_api_v2_sessions_session_id_get(
                session_id
            )
        prev_head_ip = ip_resp.result.head_node_ip

        cluster_config = up_response.cluster_config

        if cluster_config["provider"].get("type") != "kubernetes":
            cluster_config = configure_for_session(
                session_id,
                project_definition.root,
                api_client=self.api_client,
                disable_project_sync=disable_sync,
                _DO_NOT_USE_RAY_UP_ONLY_cluster_config=cluster_config,
            )

        install_anyscale_hooks(cluster_config)

        with tempfile.NamedTemporaryFile(mode="w") as config_file:
            renamed_config = copy.deepcopy(cluster_config)
            renamed_config["cluster_name"] = up_response.cluster_name
            json.dump(renamed_config, config_file)
            config_file.flush()
            try:
                # Use the bundled Autoscaler by directly calling the scripts file
                command = [
                    sys.executable,
                    autoscaler_scripts.__file__,
                    "up",
                    "-vvvv" if verbose else "",
                    config_file.name,
                    "--no-restart" if no_restart else "",
                    "--restart-only" if restart_only else "",
                    "--yes",
                ]
                command_lst = [c for c in command if c]
                if max_workers:
                    command_lst.extend(["--max-workers", f"{max_workers}"])
                if min_workers:
                    command_lst.extend(["--min-workers", f"{min_workers}"])

                # Set OS enviornment variables temporarily.
                aws_credentials = get_credentials_as_env_vars_from_cluster_config(
                    cluster_config
                )

                env = {
                    **ANYSCALE_ENV,
                    **aws_credentials,
                }

                proc = subprocess.Popen(
                    command_lst,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    env=env,
                )
                startup_log = []
                while proc.stdout:
                    line = proc.stdout.readline().decode()
                    if not line:
                        break
                    print(line, end="")
                    startup_log.append(line)
                startup_log_str = "".join(startup_log)
                proc.communicate()

                ray_up_succeeded = proc.returncode == 0

                curr_head_ip = get_head_node_ip(renamed_config)

                if curr_head_ip != prev_head_ip and ray_up_succeeded:
                    print("Setting up Jupyter lab, Ray dashboard, and autopush ...")

                self.api_client.session_finish_up_api_v2_sessions_session_id_finish_up_post(
                    session_id=session_id,
                    session_finish_up_options=SessionFinishUpOptions(
                        startup_log=startup_log_str,
                        new_session=curr_head_ip != prev_head_ip,
                        head_node_ip=curr_head_ip,
                        cluster_launch_errored=not ray_up_succeeded,
                    ),
                )
                # NOTE This call will raise an error if the session is in any of the following
                # states (StartupErrored/Stopped/Terminated).
                wait_for_session_start(project_id, session_name, self.api_client)
                url = get_endpoint(f"/projects/{project_id}")
                print(f"Session {session_name} started. View at {url}")
            except Exception as e:
                raise click.ClickException("{}\nSession startup failed.".format(e))

    def up_internal(
        self,
        session_name: Optional[str],
        config: Optional[str],
        min_workers: Optional[int],
        max_workers: Optional[int],
        no_restart: bool,
        restart_only: bool,
        disable_sync: bool,
        cloud_id: Optional[str],
        cloud_name: Optional[str],
        idle_timeout: int,
        verbose: bool = False,
    ) -> None:
        # TODO(ilr) Split this into ~3 functions, look at jbai's comments on #1469
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)

        if no_restart and restart_only:
            raise click.ClickException(
                "Cannot set both --restart-only and --no-restart at the same time."
            )

        if not session_name:
            session_name = str(
                self.api_client.get_project_default_session_name_api_v2_projects_project_id_default_session_name_get(
                    project_id=project_id,
                ).result.name
            )
        else:
            session_name = slugify(session_name)

        session_list = self.api_client.list_sessions_api_v2_sessions_get(
            project_id=project_id, active_only=False, name=session_name
        ).results

        session_exists = len(session_list) > 0

        if session_exists and not config:
            # Get cluster config stored in DB from existing session
            session = session_list[0]
            resp = self.api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get(
                session.id
            )
            cluster_config = yaml.safe_load(resp.result.config_with_defaults)
        elif not session_exists and not config:
            # Get cluster config stored in DB from project
            resp = self.api_client.get_project_latest_cluster_config_api_v2_projects_project_id_latest_cluster_config_get(
                project_id=project_id
            )
            cluster_config = yaml.safe_load(resp.result.config_with_defaults)
        else:
            # Get cluster config from `config` argument
            config = str(config)
            if not os.path.exists(config):
                raise ValueError("Config file {} not found".format(config))
            with open(config) as f:
                cluster_config_filled = populate_session_args(f.read(), config)
                cluster_config = yaml.safe_load(cluster_config_filled)
            validate_cluster_configuration(config, cluster_config, self.api_client)

        cluster_config = _configure_min_max_workers(
            cluster_config, min_workers, max_workers
        )

        cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_id, cloud_name)
        assert cloud_id is not None, "Failed to get cloud."

        with format_api_exception(ApiException):
            up_response = self.api_client.start_empty_session_api_v2_sessions_start_empty_session_post(
                session_up_options=SessionUpOptions(
                    project_id=project_id,
                    name=session_name,
                    cluster_config={"config": json.dumps(cluster_config)},
                    cloud_id=cloud_id,
                    idle_timeout=idle_timeout,
                    restart_only=restart_only,
                    no_restart=no_restart,
                )
            ).result
        session_id = up_response.session_id
        second_update_required = up_response.second_update_required
        wait_for_session_start(project_id, session_name, self.api_client)

        # Push will sync project directory to working directory.
        self.push(
            session_name, source=None, target=None, config=None, all_nodes=False,
        )

        if second_update_required:
            with format_api_exception(ApiException):
                self.api_client.setup_and_initialize_session_api_v2_sessions_session_id_setup_and_initialize_session_post(
                    session_id=session_id,
                    setup_initialize_session_options=SetupInitializeSessionOptions(
                        project_id=project_id,
                        cluster_config={"config": json.dumps(cluster_config)},
                        restart_only=restart_only,
                        no_restart=no_restart,
                    ),
                )

            wait_for_session_start(project_id, session_name, self.api_client)

        url = get_endpoint(f"/projects/{project_id}")
        print(f"Session {session_name} started. View at {url}")

    def fork_session(
        self,
        session_name: str,
        new_session_name: str,
        project_name: Optional[str] = None,
    ) -> str:
        source_session = self._resolve_session(session_name, project_name)
        if source_session.state == SessionState.RUNNING:
            print(
                "WARNING: Session is currently running, this may take several minutes."
            )

        project_url = get_endpoint(f"/projects/{source_session.project_id}")
        print(
            f"Forking session {session_name} into {new_session_name}. You can view its progress at {project_url}"
        )
        cloned_session = self._fork_session_internal(source_session, new_session_name)
        print(f"Session {session_name} forked.")

        wait_for_session_start(
            cloned_session.project_id, cloned_session.name, self.api_client
        )
        url = get_endpoint(
            f"/projects/{cloned_session.project_id}/sessions/{cloned_session.id}"
        )
        return f"Session {cloned_session.name} started. View at {url}"

    def ssh(self, session_name: str, ssh_option: Tuple[str]) -> None:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session = get_project_session(project_id, session_name, self.api_client)

        cluster_config = get_cluster_config(session_name, self.api_client)

        with format_api_exception(ApiException):
            head_ip = self.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
                session.id
            ).result.head_ip

        ssh_user = cluster_config["auth"]["ssh_user"]
        key_path = cluster_config["auth"]["ssh_private_key"]
        container_name = get_container_name(cluster_config)

        command = (
            ["ssh"]
            + list(ssh_option)
            + ["-tt", "-i", key_path]
            + ["{}@{}".format(ssh_user, head_ip)]
            + (
                [f"docker exec -it {container_name} sh -c 'which bash && bash || sh'"]
                if container_name
                else []
            )
        )

        subprocess.run(command)  # noqa: B1

    def autopush(
        self,
        session_name: Optional[str] = None,
        verbose: bool = False,
        delete: bool = False,
        ignore_errors: bool = False,
    ) -> None:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        print(f"Active project: {project_definition.root}\n")

        session = get_project_session(project_id, session_name, self.api_client)

        wait_for_session_start(project_id, session.name, self.api_client)

        cluster_config = get_cluster_config(session_name, self.api_client)

        rsync_exclude = cluster_config.get("rsync_exclude") or []
        rsync_filter = cluster_config.get("rsync_filter") or []

        with format_api_exception(ApiException):
            head_ip = self.api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
                session.id
            ).result.head_ip
        ssh_user = cluster_config["auth"]["ssh_user"]
        ssh_private_key_path = cluster_config["auth"]["ssh_private_key"]

        ssh_command = [
            "ssh",
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "UserKnownHostsFile={}".format(os.devnull),
            "-o",
            "LogLevel=ERROR",
            "-i",
            ssh_private_key_path,
        ]
        source = project_definition.root
        target = get_working_dir(cluster_config, project_id, self.api_client)
        if bool(get_container_name(cluster_config)):
            target = get_working_dir_host_mount_location(
                ssh_command,
                ssh_user,
                head_ip,
                get_container_name(cluster_config),
                target,
            )

        print("Autopush with session {} is starting up...".format(session.name))
        with managed_autosync_session(session.id, self.api_client):
            # Performing initial full synchronization with rsync.
            first_sync = perform_sync(
                ssh_command,
                source,
                ssh_user,
                head_ip,
                target,
                delete,
                rsync_filter=rsync_filter,
                rsync_exclude=rsync_exclude,
            )  # noqa: B1

            if first_sync.returncode != 0:
                message = f"First sync failed with error code: {first_sync.returncode}."
                if ignore_errors:
                    print(
                        f'{message} Continuing to sync due to "--ignore-error" option.'
                    )
                else:
                    raise click.ClickException(
                        f"{message} If you would like to continue syncing, please use "
                        'autopush with "--ignore-error" option.'
                    )

            perform_autopush_synchronization(
                ssh_command,
                source,
                ssh_user,
                head_ip,
                target,
                delete=delete,
                rsync_filter=rsync_filter,
                rsync_exclude=rsync_exclude,
            )

    def pull(
        self,
        session_name: Optional[str] = None,
        source: Optional[str] = None,
        target: Optional[str] = None,
        config: Optional[str] = None,
    ) -> None:
        project_definition = load_project_or_throw()

        try:
            print("Collecting files from remote.")
            project_id = get_project_id(project_definition.root)
            cluster_config = get_cluster_config(session_name, self.api_client)
            directory_name = get_working_dir(
                cluster_config, project_id, self.api_client
            )
            source_directory = f"{directory_name}/"

            aws_credentials = get_credentials_as_env_vars_from_cluster_config(
                cluster_config
            )

            source = canonicalize_remote_location(cluster_config, source, project_id)

            with set_env(**aws_credentials):
                if source and target:
                    rsync(
                        cluster_config, source=source, target=target, down=True,
                    )
                elif source or target:
                    raise click.ClickException(
                        "Source and target are not both specified. Please either specify both or neither."
                    )
                else:
                    rsync(
                        cluster_config,
                        source=source_directory,
                        target=project_definition.root,
                        down=True,
                    )

            if config:
                session = get_project_session(project_id, session_name, self.api_client)
                with format_api_exception(ApiException):
                    resp = self.api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get(
                        session.id
                    )
                cluster_config = yaml.safe_load(resp.result.config_with_defaults)
                with open(config, "w") as f:
                    yaml.dump(cluster_config, f, default_flow_style=False)

            print("Pull completed.")

        except Exception as e:
            print(e)
            raise click.ClickException(e)  # type: ignore

    def push(
        self,
        session_name: str,
        source: Optional[str],
        target: Optional[str],
        config: Optional[str],
        all_nodes: bool,
    ) -> None:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session = get_project_session(project_id, session_name, self.api_client)
        session_name = session.name

        cluster_config = get_cluster_config(session_name, self.api_client)
        target = canonicalize_remote_location(cluster_config, target, project_id)

        aws_credentials = get_credentials_as_env_vars_from_cluster_config(
            cluster_config
        )

        with set_env(**aws_credentials):
            if source and target:
                rsync(
                    cluster_config, source=source, target=target, down=False,
                )
            elif source or target:
                raise click.ClickException(
                    "Source and target are not both specified. Please either specify both or neither."
                )
            else:
                rsync(
                    cluster_config, source=None, target=None, down=False,
                )

        if config:
            validate_cluster_configuration(config, api_instance=self.api_client)
            print("Updating session with {}".format(config))
            self.up(
                session_name=session_name,
                config=None,
                build_id=None,
                compute_template_id=None,
                min_workers=None,
                max_workers=None,
                no_restart=False,
                restart_only=False,
                disable_sync=True,
                cloud_id=session.cloud_id,
                cloud_name=None,
                idle_timeout=None,
                yes=True,
            )
        url = get_endpoint(f"/projects/{project_id}")
        print(f"Pushed to session {session_name}. View at {url}")

    # Helpers

    def _resolve_session(
        self, session_name: str, project_name: Optional[str] = None
    ) -> Session:
        """
        Resolves a session by name.
        This is distinct from  `anyscale.project.get_project_session` because:
        1. we rely on project names instead of ids
        2. we allow non-active sessions to be resolved

        Raises an exception if the session does not exist.

        Params
        session_name - name of the session
        project_name - optional project name that the session is in;
                       if absent, we use the workspace's project
        """
        if project_name:
            with format_api_exception(ApiException):
                projects = self.api_client.list_projects_api_v2_projects_get().results
            project = next(
                project for project in projects if project.name == project_name
            )
            project_id = project.id
        else:
            project_definition = load_project_or_throw()
            project_id = get_project_id(project_definition.root)

        with format_api_exception(ApiException):
            sessions_list = self.api_client.list_sessions_api_v2_sessions_get(
                project_id, name=session_name
            ).results

        if len(sessions_list) == 0:
            raise click.ClickException(
                f"No session found with name {session_name} in project {project_id}"
            )

        return sessions_list[0]

    def _fork_session_internal(
        self, source_session: Session, new_session_name: str
    ) -> Session:
        """
        Creates a clone of the source session.
        Raises an exception if cloning fails.

        Params
        source_session - session to clone
        new_session_name - name of the new session
        """
        with format_api_exception(ApiException):
            self.api_client.fork_session_api_v2_sessions_session_id_fork_post(
                session_id=source_session.id,
                create_session_from_snapshot_options=CreateSessionFromSnapshotOptions(
                    project_id=source_session.project_id, name=new_session_name
                ),
                # forking an active session can take up to 180s
                _request_timeout=300000,
            )

        with format_api_exception(ApiException):
            forked_sessions_list = self.api_client.list_sessions_api_v2_sessions_get(
                project_id=source_session.project_id, name=new_session_name
            ).results

        if len(forked_sessions_list) == 0:
            raise click.ClickException("Unable to fork session.")

        return forked_sessions_list[0]



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SocketMessageSchemas(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'project_create_message': 'ProjectCreateMessage',
        'project_delete_message': 'ProjectDeleteMessage',
        'project_patch_message': 'ProjectPatchMessage',
        'session_autosync_sessions_update_message': 'SessionAutosyncSessionsUpdateMessage',
        'session_execute_message': 'SessionExecuteMessage',
        'session_kill_command_message': 'SessionKillCommandMessage',
        'session_stop_message': 'SessionStopMessage',
        'session_finish_command_message': 'SessionFinishCommandMessage',
        'snapshot_create_message': 'SnapshotCreateMessage',
        'snapshot_delete_message': 'SnapshotDeleteMessage',
        'snapshot_patch_message': 'SnapshotPatchMessage'
    }

    attribute_map = {
        'project_create_message': 'ProjectCreateMessage',
        'project_delete_message': 'ProjectDeleteMessage',
        'project_patch_message': 'ProjectPatchMessage',
        'session_autosync_sessions_update_message': 'SessionAutosyncSessionsUpdateMessage',
        'session_execute_message': 'SessionExecuteMessage',
        'session_kill_command_message': 'SessionKillCommandMessage',
        'session_stop_message': 'SessionStopMessage',
        'session_finish_command_message': 'SessionFinishCommandMessage',
        'snapshot_create_message': 'SnapshotCreateMessage',
        'snapshot_delete_message': 'SnapshotDeleteMessage',
        'snapshot_patch_message': 'SnapshotPatchMessage'
    }

    def __init__(self, project_create_message=None, project_delete_message=None, project_patch_message=None, session_autosync_sessions_update_message=None, session_execute_message=None, session_kill_command_message=None, session_stop_message=None, session_finish_command_message=None, snapshot_create_message=None, snapshot_delete_message=None, snapshot_patch_message=None, local_vars_configuration=None):  # noqa: E501
        """SocketMessageSchemas - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._project_create_message = None
        self._project_delete_message = None
        self._project_patch_message = None
        self._session_autosync_sessions_update_message = None
        self._session_execute_message = None
        self._session_kill_command_message = None
        self._session_stop_message = None
        self._session_finish_command_message = None
        self._snapshot_create_message = None
        self._snapshot_delete_message = None
        self._snapshot_patch_message = None
        self.discriminator = None

        self.project_create_message = project_create_message
        self.project_delete_message = project_delete_message
        self.project_patch_message = project_patch_message
        self.session_autosync_sessions_update_message = session_autosync_sessions_update_message
        self.session_execute_message = session_execute_message
        self.session_kill_command_message = session_kill_command_message
        self.session_stop_message = session_stop_message
        self.session_finish_command_message = session_finish_command_message
        self.snapshot_create_message = snapshot_create_message
        self.snapshot_delete_message = snapshot_delete_message
        self.snapshot_patch_message = snapshot_patch_message

    @property
    def project_create_message(self):
        """Gets the project_create_message of this SocketMessageSchemas.  # noqa: E501


        :return: The project_create_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: ProjectCreateMessage
        """
        return self._project_create_message

    @project_create_message.setter
    def project_create_message(self, project_create_message):
        """Sets the project_create_message of this SocketMessageSchemas.


        :param project_create_message: The project_create_message of this SocketMessageSchemas.  # noqa: E501
        :type: ProjectCreateMessage
        """
        if self.local_vars_configuration.client_side_validation and project_create_message is None:  # noqa: E501
            raise ValueError("Invalid value for `project_create_message`, must not be `None`")  # noqa: E501

        self._project_create_message = project_create_message

    @property
    def project_delete_message(self):
        """Gets the project_delete_message of this SocketMessageSchemas.  # noqa: E501


        :return: The project_delete_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: ProjectDeleteMessage
        """
        return self._project_delete_message

    @project_delete_message.setter
    def project_delete_message(self, project_delete_message):
        """Sets the project_delete_message of this SocketMessageSchemas.


        :param project_delete_message: The project_delete_message of this SocketMessageSchemas.  # noqa: E501
        :type: ProjectDeleteMessage
        """
        if self.local_vars_configuration.client_side_validation and project_delete_message is None:  # noqa: E501
            raise ValueError("Invalid value for `project_delete_message`, must not be `None`")  # noqa: E501

        self._project_delete_message = project_delete_message

    @property
    def project_patch_message(self):
        """Gets the project_patch_message of this SocketMessageSchemas.  # noqa: E501


        :return: The project_patch_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: ProjectPatchMessage
        """
        return self._project_patch_message

    @project_patch_message.setter
    def project_patch_message(self, project_patch_message):
        """Sets the project_patch_message of this SocketMessageSchemas.


        :param project_patch_message: The project_patch_message of this SocketMessageSchemas.  # noqa: E501
        :type: ProjectPatchMessage
        """
        if self.local_vars_configuration.client_side_validation and project_patch_message is None:  # noqa: E501
            raise ValueError("Invalid value for `project_patch_message`, must not be `None`")  # noqa: E501

        self._project_patch_message = project_patch_message

    @property
    def session_autosync_sessions_update_message(self):
        """Gets the session_autosync_sessions_update_message of this SocketMessageSchemas.  # noqa: E501


        :return: The session_autosync_sessions_update_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SessionAutosyncSessionsUpdateMessage
        """
        return self._session_autosync_sessions_update_message

    @session_autosync_sessions_update_message.setter
    def session_autosync_sessions_update_message(self, session_autosync_sessions_update_message):
        """Sets the session_autosync_sessions_update_message of this SocketMessageSchemas.


        :param session_autosync_sessions_update_message: The session_autosync_sessions_update_message of this SocketMessageSchemas.  # noqa: E501
        :type: SessionAutosyncSessionsUpdateMessage
        """
        if self.local_vars_configuration.client_side_validation and session_autosync_sessions_update_message is None:  # noqa: E501
            raise ValueError("Invalid value for `session_autosync_sessions_update_message`, must not be `None`")  # noqa: E501

        self._session_autosync_sessions_update_message = session_autosync_sessions_update_message

    @property
    def session_execute_message(self):
        """Gets the session_execute_message of this SocketMessageSchemas.  # noqa: E501


        :return: The session_execute_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SessionExecuteMessage
        """
        return self._session_execute_message

    @session_execute_message.setter
    def session_execute_message(self, session_execute_message):
        """Sets the session_execute_message of this SocketMessageSchemas.


        :param session_execute_message: The session_execute_message of this SocketMessageSchemas.  # noqa: E501
        :type: SessionExecuteMessage
        """
        if self.local_vars_configuration.client_side_validation and session_execute_message is None:  # noqa: E501
            raise ValueError("Invalid value for `session_execute_message`, must not be `None`")  # noqa: E501

        self._session_execute_message = session_execute_message

    @property
    def session_kill_command_message(self):
        """Gets the session_kill_command_message of this SocketMessageSchemas.  # noqa: E501


        :return: The session_kill_command_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SessionKillCommandMessage
        """
        return self._session_kill_command_message

    @session_kill_command_message.setter
    def session_kill_command_message(self, session_kill_command_message):
        """Sets the session_kill_command_message of this SocketMessageSchemas.


        :param session_kill_command_message: The session_kill_command_message of this SocketMessageSchemas.  # noqa: E501
        :type: SessionKillCommandMessage
        """
        if self.local_vars_configuration.client_side_validation and session_kill_command_message is None:  # noqa: E501
            raise ValueError("Invalid value for `session_kill_command_message`, must not be `None`")  # noqa: E501

        self._session_kill_command_message = session_kill_command_message

    @property
    def session_stop_message(self):
        """Gets the session_stop_message of this SocketMessageSchemas.  # noqa: E501


        :return: The session_stop_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SessionStopMessage
        """
        return self._session_stop_message

    @session_stop_message.setter
    def session_stop_message(self, session_stop_message):
        """Sets the session_stop_message of this SocketMessageSchemas.


        :param session_stop_message: The session_stop_message of this SocketMessageSchemas.  # noqa: E501
        :type: SessionStopMessage
        """
        if self.local_vars_configuration.client_side_validation and session_stop_message is None:  # noqa: E501
            raise ValueError("Invalid value for `session_stop_message`, must not be `None`")  # noqa: E501

        self._session_stop_message = session_stop_message

    @property
    def session_finish_command_message(self):
        """Gets the session_finish_command_message of this SocketMessageSchemas.  # noqa: E501


        :return: The session_finish_command_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SessionFinishCommandMessage
        """
        return self._session_finish_command_message

    @session_finish_command_message.setter
    def session_finish_command_message(self, session_finish_command_message):
        """Sets the session_finish_command_message of this SocketMessageSchemas.


        :param session_finish_command_message: The session_finish_command_message of this SocketMessageSchemas.  # noqa: E501
        :type: SessionFinishCommandMessage
        """
        if self.local_vars_configuration.client_side_validation and session_finish_command_message is None:  # noqa: E501
            raise ValueError("Invalid value for `session_finish_command_message`, must not be `None`")  # noqa: E501

        self._session_finish_command_message = session_finish_command_message

    @property
    def snapshot_create_message(self):
        """Gets the snapshot_create_message of this SocketMessageSchemas.  # noqa: E501


        :return: The snapshot_create_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SnapshotCreateMessage
        """
        return self._snapshot_create_message

    @snapshot_create_message.setter
    def snapshot_create_message(self, snapshot_create_message):
        """Sets the snapshot_create_message of this SocketMessageSchemas.


        :param snapshot_create_message: The snapshot_create_message of this SocketMessageSchemas.  # noqa: E501
        :type: SnapshotCreateMessage
        """
        if self.local_vars_configuration.client_side_validation and snapshot_create_message is None:  # noqa: E501
            raise ValueError("Invalid value for `snapshot_create_message`, must not be `None`")  # noqa: E501

        self._snapshot_create_message = snapshot_create_message

    @property
    def snapshot_delete_message(self):
        """Gets the snapshot_delete_message of this SocketMessageSchemas.  # noqa: E501


        :return: The snapshot_delete_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SnapshotDeleteMessage
        """
        return self._snapshot_delete_message

    @snapshot_delete_message.setter
    def snapshot_delete_message(self, snapshot_delete_message):
        """Sets the snapshot_delete_message of this SocketMessageSchemas.


        :param snapshot_delete_message: The snapshot_delete_message of this SocketMessageSchemas.  # noqa: E501
        :type: SnapshotDeleteMessage
        """
        if self.local_vars_configuration.client_side_validation and snapshot_delete_message is None:  # noqa: E501
            raise ValueError("Invalid value for `snapshot_delete_message`, must not be `None`")  # noqa: E501

        self._snapshot_delete_message = snapshot_delete_message

    @property
    def snapshot_patch_message(self):
        """Gets the snapshot_patch_message of this SocketMessageSchemas.  # noqa: E501


        :return: The snapshot_patch_message of this SocketMessageSchemas.  # noqa: E501
        :rtype: SnapshotPatchMessage
        """
        return self._snapshot_patch_message

    @snapshot_patch_message.setter
    def snapshot_patch_message(self, snapshot_patch_message):
        """Sets the snapshot_patch_message of this SocketMessageSchemas.


        :param snapshot_patch_message: The snapshot_patch_message of this SocketMessageSchemas.  # noqa: E501
        :type: SnapshotPatchMessage
        """
        if self.local_vars_configuration.client_side_validation and snapshot_patch_message is None:  # noqa: E501
            raise ValueError("Invalid value for `snapshot_patch_message`, must not be `None`")  # noqa: E501

        self._snapshot_patch_message = snapshot_patch_message

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SocketMessageSchemas):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SocketMessageSchemas):
            return True

        return self.to_dict() != other.to_dict()



import copy
import logging
import os
import threading
from types import ModuleType
from typing import Any, Dict, List, Optional

import docker
import filelock
from ray.autoscaler.command_runner import CommandRunnerInterface
from ray.autoscaler.node_provider import NodeProvider
from ray.autoscaler.tags import NODE_KIND_HEAD, TAG_RAY_CLUSTER_NAME, TAG_RAY_NODE_KIND
import yaml


from anyscale.autoscaler.docker_local import client
from anyscale.autoscaler.docker_local.command_runner import DockerLocalCommandRunner

logger = logging.getLogger(__name__)
log_prefix = "DockerLocalNodeProvider: "

# Silence filelock logging.
logging.getLogger("filelock").setLevel(logging.ERROR)


def container_status(node_id: str) -> str:
    try:
        container = client().containers.get(node_id)
        return container.status  # type: ignore
    except docker.errors.NotFound:
        return "not found"


def stop_container(node_id: str) -> None:
    try:
        container = client().containers.get(node_id)
        container.stop()
    except docker.errors.NotFound:
        logger.warn(
            log_prefix + "Tried to stop container with name"
            f"{node_id}, but no container with that name was found."
        )


def get_docker_run_kwargs(
    node_config: Dict[str, Any], cluster_name: str
) -> Dict[str, Any]:
    """Converts node_config into a dict of keyword arguments to be passed to
    docker run in DockerLocalNodeProvider.create_node().

    Prepares "ray-cluster-name" label for the container.
    Converts generate_name field of node_config into a container name by
    appending a UUID.
    """
    kwargs = copy.deepcopy(node_config)
    return kwargs


# TODO (Dmitri) : Get rid of this  when we add logic to run autoscaler
# outside of container.
DOCKER_SOCKET = "/var/run/docker.sock"


def get_head_mounts(*paths: str) -> Dict[str, Any]:
    head_mounts = {}
    for path in paths:
        head_mounts.update({path: {"bind": path, "mount": "rw"}})
    return head_mounts


TAGS_PATH_TEMPLATE = "/tmp/ray_docker_local_cluster-{}.tags"
LOCK_PATH_TEMPLATE = "/tmp/ray_docker_local_cluster-{}.lock"


class DockerLocalNodeTags:
    """Helper class used by DockerLocalNodeProvider to set and get node tags.

    Maintains a temp file with {container name: tags} mapping in yaml format.
    """

    def __init__(self, cluster_name: str) -> None:
        self.tag_path = TAGS_PATH_TEMPLATE.format(cluster_name)
        self.lock_path = LOCK_PATH_TEMPLATE.format(cluster_name)

        # Locks file operations between threads in a single process.
        self.lock = threading.Lock()
        # Locks file operations between processes.
        self.file_lock = filelock.FileLock(self.lock_path)

        with self.lock, self.file_lock:
            if not os.path.exists(self.tag_path):
                self._write({})

    def get(self, node_id: str) -> Dict[str, str]:
        with self.lock, self.file_lock:
            try:
                return self._read()[node_id]
            except Exception:
                # Avoids a transient error during cluster shutdown.
                # TODO (dmitri): Fix for real!
                logger.error("Failed to read node tags!" "Returning an empty dict.")
                return {}

    def set(self, node_id: str, tags: Dict[str, str]) -> None:
        with self.lock, self.file_lock:
            all_tags = self._read()
            if node_id not in all_tags:
                all_tags[node_id] = {}
            all_tags[node_id].update(tags)
            self._write(all_tags)

    def delete(self, node_id: str) -> None:
        with self.lock, self.file_lock:
            all_tags = self._read()
            try:
                all_tags.pop(node_id)
            except KeyError:
                logger.warning(
                    log_prefix + "Tried to delete tags of node with "
                    f"id {node_id}, but the node's tags are already gone."
                )
            self._write(all_tags)

    def _read(self) -> Dict[str, Dict[str, str]]:
        with open(self.tag_path, "r") as tag_file:
            return yaml.safe_load(tag_file)  # type: ignore

    def _write(self, all_tags: Dict[str, Dict[str, str]]) -> None:
        with open(self.tag_path, "w") as tag_file:
            yaml.dump(all_tags, tag_file, default_flow_style=False)


class DockerLocalNodeProvider(NodeProvider):  # type: ignore
    """NodeProvider for local Ray experiments and tests.

    Provides a Docker container on local machine for each Ray node.
    """

    def __init__(self, provider_config: Dict[str, Any], cluster_name: str) -> None:
        NodeProvider.__init__(self, provider_config, cluster_name)
        self.tag_manager = DockerLocalNodeTags(cluster_name)

    def non_terminated_nodes(self, tag_filters: Dict[str, str]) -> List[str]:
        """Return a list of node ids filtered by the specified tags dict.

        This list must not include terminated nodes. For performance reasons,
        providers are allowed to cache the result of a call to nodes() to
        serve single-node queries (e.g. is_running(node_id)). This means that
        nodes() must be called again to refresh results.

        Examples:
            >>> provider.non_terminated_nodes({TAG_RAY_NODE_KIND: "worker"})
            ["node-1", "node-2"]
        """
        filters = {"label": f"{TAG_RAY_CLUSTER_NAME}={self.cluster_name}"}
        all_node_ids = [
            container.name for container in client().containers.list(filters=filters)
        ]
        return [
            node_id
            for node_id in all_node_ids
            if tag_filters.items() <= self.node_tags(node_id).items()
        ]

    def is_running(self, node_id: str) -> bool:
        """Return whether the specified node is running."""
        return container_status(node_id) == "running"

    def is_terminated(self, node_id: str) -> bool:
        """Return whether the specified node is terminated."""
        return container_status(node_id) != "running"

    def node_tags(self, node_id: str) -> Dict[str, str]:
        """Returns the tags of the given node (string dict)."""
        return self.tag_manager.get(node_id)

    def internal_ip(self, node_id: str) -> str:
        """Returns the internal ip (Ray ip) of the given node."""
        container_info = client().api.inspect_container(node_id)
        return container_info["NetworkSettings"]["Networks"]["bridge"]["IPAddress"]  # type: ignore

    def create_node(
        self, node_config: Dict[str, Any], tags: Dict[str, str], count: int
    ) -> None:
        """Creates a number of nodes within the namespace."""

        # Add cluster_name label to container run arguments.
        docker_run_kwargs = copy.deepcopy(node_config)
        cluster_label = {TAG_RAY_CLUSTER_NAME: self.cluster_name}
        docker_run_kwargs.update({"labels": cluster_label})

        # TODO (Dmitri) : Get rid of this when we add logic to run autoscaler
        # outside of container.
        if tags[TAG_RAY_NODE_KIND] == NODE_KIND_HEAD:
            head_mounts = get_head_mounts(
                DOCKER_SOCKET, self.tag_manager.tag_path, self.tag_manager.lock_path
            )
            if "volumes" not in docker_run_kwargs:
                docker_run_kwargs["volumes"] = {}
            docker_run_kwargs["volumes"].update(head_mounts)

        for _ in range(count):
            container = client().containers.run(**docker_run_kwargs)
            self.set_node_tags(container.name, tags)
            logger.info(log_prefix + f"Created node with id '{container.name}'")

    def set_node_tags(self, node_id: str, tags: Dict[str, str]) -> None:
        """Sets the tag values (string dict) for the specified node."""
        self.tag_manager.set(node_id, tags)

    def terminate_node(self, node_id: str) -> None:
        """Terminates the specified node."""
        logger.info(
            log_prefix + "NodeProvider: " "{}: Terminating node".format(node_id)
        )
        stop_container(node_id)
        self.tag_manager.delete(node_id)

    def terminate_nodes(self, node_ids: List[str]) -> None:
        """Terminates a set of nodes. May be overridden with a batch method."""
        for node_id in node_ids:
            self.terminate_node(node_id)

    @staticmethod
    def bootstrap_config(cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Bootstraps the cluster config by adding env defaults if needed."""
        return cluster_config

    def get_command_runner(
        self,
        log_prefix: str,
        node_id: str,
        auth_config: Dict[str, Any],
        cluster_name: str,
        process_runner: ModuleType,
        use_internal_ip: bool,
        docker_config: Optional[Dict[str, Any]] = None,
    ) -> CommandRunnerInterface:
        return DockerLocalCommandRunner(log_prefix, node_id, process_runner)

    def prepare_for_head_node(self, cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        """Returns a new cluster config with custom configs for head node."""
        return cluster_config

    @staticmethod
    def fillout_available_node_types_resources(
        cluster_config: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Fills out missing "resources" field for available_node_types."""
        return cluster_config



# coding: utf-8

# flake8: noqa

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from anyscale_client.api.default_api import DefaultApi

# import ApiClient
from anyscale_client.api_client import ApiClient
from anyscale_client.configuration import Configuration
from anyscale_client.exceptions import OpenApiException
from anyscale_client.exceptions import ApiTypeError
from anyscale_client.exceptions import ApiValueError
from anyscale_client.exceptions import ApiKeyError
from anyscale_client.exceptions import ApiException
# import models into sdk package
from anyscale_client.models.aws_credentials import AWSCredentials
from anyscale_client.models.anyscaled_config import AnyscaledConfig
from anyscale_client.models.anyscaledconfig_response import AnyscaledconfigResponse
from anyscale_client.models.application_template import ApplicationTemplate
from anyscale_client.models.applicationtemplate_list_response import ApplicationtemplateListResponse
from anyscale_client.models.applicationtemplate_response import ApplicationtemplateResponse
from anyscale_client.models.awscredentials_response import AwscredentialsResponse
from anyscale_client.models.build import Build
from anyscale_client.models.build_list_response import BuildListResponse
from anyscale_client.models.build_log_response import BuildLogResponse
from anyscale_client.models.build_response import BuildResponse
from anyscale_client.models.build_status import BuildStatus
from anyscale_client.models.buildlogresponse_response import BuildlogresponseResponse
from anyscale_client.models.compute_template import ComputeTemplate
from anyscale_client.models.compute_template_query import ComputeTemplateQuery
from anyscale_client.models.computetemplate_list_response import ComputetemplateListResponse
from anyscale_client.models.computetemplate_response import ComputetemplateResponse
from anyscale_client.models.create_application_template import CreateApplicationTemplate
from anyscale_client.models.create_build import CreateBuild
from anyscale_client.models.create_compute_template import CreateComputeTemplate
from anyscale_client.models.create_project import CreateProject
from anyscale_client.models.create_response import CreateResponse
from anyscale_client.models.create_session import CreateSession
from anyscale_client.models.create_session_command import CreateSessionCommand
from anyscale_client.models.http_validation_error import HTTPValidationError
from anyscale_client.models.list_response_metadata import ListResponseMetadata
from anyscale_client.models.operation_error import OperationError
from anyscale_client.models.operation_progress import OperationProgress
from anyscale_client.models.operation_results import OperationResults
from anyscale_client.models.page_query import PageQuery
from anyscale_client.models.project import Project
from anyscale_client.models.project_list_response import ProjectListResponse
from anyscale_client.models.project_response import ProjectResponse
from anyscale_client.models.projects_query import ProjectsQuery
from anyscale_client.models.session import Session
from anyscale_client.models.session_command import SessionCommand
from anyscale_client.models.session_list_response import SessionListResponse
from anyscale_client.models.session_operation import SessionOperation
from anyscale_client.models.session_operation_type import SessionOperationType
from anyscale_client.models.session_response import SessionResponse
from anyscale_client.models.session_state import SessionState
from anyscale_client.models.sessioncommand_list_response import SessioncommandListResponse
from anyscale_client.models.sessioncommand_response import SessioncommandResponse
from anyscale_client.models.sessionoperation_response import SessionoperationResponse
from anyscale_client.models.sessions_query import SessionsQuery
from anyscale_client.models.start_session_options import StartSessionOptions
from anyscale_client.models.stop_session_options import StopSessionOptions
from anyscale_client.models.terminate_session_options import TerminateSessionOptions
from anyscale_client.models.text_query import TextQuery
from anyscale_client.models.update_application_template import UpdateApplicationTemplate
from anyscale_client.models.update_compute_template import UpdateComputeTemplate
from anyscale_client.models.update_project import UpdateProject
from anyscale_client.models.update_session import UpdateSession
from anyscale_client.models.validation_error import ValidationError




import os
import subprocess
from typing import Any, List, Optional, Union

import ray
import yaml

from anyscale.credentials import load_credentials
import anyscale.project
from anyscale.sdk.anyscale_client.sdk import AnyscaleSDK  # type: ignore

# The default project directory to use, if no project is specified.
DEFAULT_SCRATCH_DIR = "~/anyscale-connect-scratch"


class SessionBuilder:
    """This class lets you set session options and connect to Anyscale.

    This feature is ***EXPERIMENTAL***.

    It should not be constructed directly, but instead via anyscale.* methods
    exported at the package level.

    Examples:
        >>> # Raw client, creates new session on behalf of user
        >>> anyscale.connect()

        >>> # Get or create a named session
        >>> anyscale
        ...   .session("my_named_session")
        ...   .connect()

        >>> # Specify a previously created env / app template
        >>> anyscale
        ...   .app_env("prev_created_env:v2")
        ...   .autosuspend(hours=2)
        ...   .connect()

        >>> # Create new session from local env / from scratch
        >>> anyscale
        ...   .project_dir("~/dev/my-project-folder")
        ...   .base_docker_image("anyscale/ray-ml:latest")
        ...   .require("~/dev/my-project-folder/requirements.txt")
        ...   .connect()

        >>> # Ray client connect is setup automatically
        >>> @ray.remote
        ... def my_func(value):
        ...   return value ** 2

        >>> # Remote functions are executed in the Anyscale session
        >>> print(ray.get([my_func.remote(x) for x in range(5)]))
        >>> [0, 1, 4, 9, 16]
    """

    def __init__(
        self,
        scratch_dir: str = DEFAULT_SCRATCH_DIR,
        anyscale_sdk: AnyscaleSDK = None,
        subprocess: Any = subprocess,
        ray: Any = ray,
    ) -> None:
        self._anyscale_sdk: Any = None
        if anyscale_sdk:
            self._anyscale_sdk = anyscale_sdk
        else:
            self._anyscale_sdk = AnyscaleSDK(load_credentials())
        self._ray: Any = ray
        self._subprocess: Any = subprocess
        self._scratch_dir: str = scratch_dir
        self._project_dir: Optional[str] = None
        self._session_name: str = "session-1"
        self._connected: bool = False

    def connect(self) -> None:
        """Connect to Anyscale using previously specified options.

        Examples:
            >>> anyscale.connect()
        """

        assert not self._connected, "Cannot call connect() twice."

        # Autodetect or create a scratch project.
        if self._project_dir is None:
            self._project_dir = anyscale.project.find_project_root(os.getcwd())
        if self._project_dir:
            self._ensure_project_setup_at_dir(self._project_dir)
        else:
            self._project_dir = self._get_or_create_scratch_project()
        proj_def = anyscale.project.ProjectDefinition(self._project_dir)
        project_id = anyscale.project.get_project_id(proj_def.root)
        print("Using project dir", proj_def.root)

        # TODO(ekl): generate a serverless compute configuraton here.
        cluster_yaml = yaml.safe_load(anyscale.project.CLUSTER_YAML_TEMPLATE)
        cluster_yaml["setup_commands"] = [
            "pip install -U https://s3-us-west-2.amazonaws.com/ray-wheels/"
            "latest/ray-1.2.0.dev0-cp37-cp37m-manylinux2014_x86_64.whl"
        ]
        cluster_yaml["head_start_ray_commands"] = [
            "ray stop",
            "ulimit -n 65536; ray start --head --port=6379 "
            "--object-manager-port=8076 "
            "--ray-client-server-port=10001 "
            "--autoscaling-config=~/ray_bootstrap_config.yaml",
        ]

        with open(os.path.join(self._project_dir, "session-default.yaml"), "w+") as f:
            f.write(yaml.dump(cluster_yaml))

        # TODO(ekl): better logging here
        self._subprocess.check_call(
            [
                "anyscale",
                "up",
                "--cloud-name",
                "anyscale_default_cloud",
                self._session_name,
            ],
            cwd=self._project_dir,
        )

        query = self._anyscale_sdk.search_sessions(project_id=project_id)
        session_found = None
        for session in query.results:
            if session.name == self._session_name:
                session_found = session
                break
        if not session_found:
            raise RuntimeError(
                "Failed to locate session: {}".format(self._session_name)
            )

        full_url = session_found.service_proxy_url
        # like "session-fqsx0p3pzfna71xxxxxxx.anyscaleuserdata.com:8081"
        session_url = full_url.split("/")[2].lower() + ":8081"
        # like "8218b528-7363-4d04-8358-57936cdxxxxx"
        auth_token = full_url.split("?token=")[1].split("&")[0]

        metadata = [("cookie", "anyscale-token=" + auth_token), ("port", "10001")]
        print("Connecting to Ray", session_url, metadata)
        self._ray.util.connect(session_url, secure=False, metadata=metadata)
        print("Checking connection...")

        def func() -> str:
            return "Connected!"

        f_remote = self._ray.remote(func)

        print(self._ray.get(f_remote.remote()))
        self._connected = True

    def project_dir(self, local_dir: str) -> "SessionBuilder":
        """Set the project directory.

        This sets the project code directory that will be synced to all nodes
        in the cluster as required by Ray. If not specified, the project
        directory will be autodetected based on the current working directory.
        If no Anyscale project is found, a "scratch" project will be used.

        Args:
            local_dir (str): path to the project directory.

        Examples:
            >>> anyscale.project_dir("~/my-proj-dir").connect()
        """
        self._project_dir = os.path.abspath(os.path.expanduser(local_dir))
        return self

    def session(self, session_name: str, update: bool = True) -> "SessionBuilder":
        """Set a fixed session name.

        By default, Anyscale connect will pick an idle session compatible
        with the connect parameters, creating a new session if no compatible
        idle sessions. Setting a fixed session name will force connecting to
        the given named session, creating it if it doesn't exist.

        Args:
            session_name (str): fixed name of the session.
            update (bool): whether to update session configurations when
                connecting to an existing session. Note that this may restart
                the Ray runtime.

        Examples:
            >>> anyscale.session("prod_deployment", update=True).connect()
        """
        if not update:
            raise NotImplementedError("TODO implement no update connect")
        self._session_name = session_name
        return self

    def base_docker_image(self, image_name: str) -> "SessionBuilder":
        """Set the docker image to use for the session.

        Args:
            image_name (str): docker image name.

        Examples:
            >>> anyscale.base_docker_image("anyscale/ray-ml:latest").connect()
        """
        raise NotImplementedError()

    def require(self, requirements: Union[str, List[str]]) -> "SessionBuilder":
        """Set the Python requirements for the session.

        Args:
            requirements: either be a list of pip library specifications, or
            the path to a requirements.txt file.

        Examples:
            >>> anyscale.require("~/proj/requirements.txt").connect()
            >>> anyscale.require(["gym", "torch>=1.4.0"]).connect()
        """
        raise NotImplementedError()

    def app_env(self, env_name: str) -> "SessionBuilder":
        """Set the Anyscale app env to use for the session.

        Args:
            env_name (str): app env name.

        Examples:
            >>> anyscale.app_env("prev_created_env:v2").connect()
        """
        raise NotImplementedError()

    def file_mount(self, *, local_dir: str, remote_dir: str) -> "SessionBuilder":
        """Add additional directories to sync up to worker nodes.

        Args:
            local_dir (str): the local directory path to mount.
            remote_dir (str): where in the remote node to mount the local dir.

        Examples:
            >>> anyscale
            ...   .file_mount(local_dir="~/data1", remote_dir="/tmp/d1")
            ...   .file_mount(local_dir="~/data2", remote_dir="/tmp/d2")
            ...   .connect()
        """
        raise NotImplementedError()

    def download_results(
        self, *, remote_dir: str, local_dir: str, autosync: bool = False
    ) -> "SessionBuilder":
        """Specify a directory to sync down from the cluster head node.

        Args:
            remote_dir (str): the remote result dir on the head node.
            local_dir (str): the local path to sync results to.
            autosync (bool): whether to sync the files continuously. By
                default, results will only be synced on job completion.

        Examples:
            >>> anyscale
            ...   .download_results(
            ...       remote_dir="~/ray_results", remote_dir="~/proj_output",
            ...       autosync=True)
            ...   .connect()
        """
        raise NotImplementedError()

    def autoshutdown(
        self, enabled: bool = True, *, hours: int = 1, minutes: int = 0,
    ) -> "SessionBuilder":
        """Configure or disable session autoshutdown behavior.

        The session will be autoshutdown after the specified time period. By
        default, sessions auto terminate after one hour of idle.

        Args:
            enabled (bool): whether autoshutdown is enabled.
            hours (int): specify idle time in hours.
            minutes (int): specify idle time in minutes. This is added to the
                idle time in hours.

        Examples:
            >>> anyscale.autoshutdown(False).connect()
            >>> anyscale.autoshutdown(hours=10).connect()
            >>> anyscale.autoshutdown(hours=1, minutes=30).connect()
        """
        raise NotImplementedError()

    def _get_or_create_scratch_project(self) -> str:
        """Get or create a scratch project, including the directory."""
        project_dir = os.path.expanduser(self._scratch_dir)
        self._ensure_project_setup_at_dir(project_dir)
        return project_dir

    def _ensure_project_setup_at_dir(self, project_dir: str) -> None:
        """Get or create an Anyscale project rooted at the given dir."""
        os.makedirs(project_dir, exist_ok=True)
        proj_name = os.path.basename(project_dir)

        # If the project yaml exists, assume we're already setup.
        project_yaml = os.path.join(project_dir, ".anyscale.yaml")
        if os.path.exists(project_yaml):
            return

        print("Creating new project for", project_dir)
        project_response = self._anyscale_sdk.create_project(
            {
                "name": proj_name,
                "description": "Automatically created by Anyscale Connect",
            }
        )
        project_id = project_response.result.id

        if not os.path.exists(project_yaml):
            with open(project_yaml, "w+") as f:
                f.write(yaml.dump({"project_id": project_id}))



from typing import List, Optional

import click

from anyscale.project import (
    get_project_id,
    get_project_session,
    load_project_or_throw,
    ProjectDefinition,
)
from anyscale.snapshot import describe_snapshot
from anyscale.util import (
    get_endpoint,
    send_json_request,
)


@click.group("snapshot", help="Commands for working with snapshot.", hidden=True)
def snapshot_cli() -> None:
    pass


def remote_snapshot(
    project_id: str,
    session_name: str,
    tags: List[str],
    project_definition: ProjectDefinition,
    description: Optional[str] = None,
) -> str:
    session = get_project_session(project_id, session_name)

    resp = send_json_request(
        "/api/v2/sessions/{session_id}/take_snapshot".format(session_id=session["id"]),
        {
            "tags": tags,
            "description": description if description else "",
        },  # noqa: E231
        method="POST",
    )
    if "id" not in resp["result"]:
        raise click.ClickException(
            "Snapshot creation of session {} failed!".format(session["name"])
        )
    snapshot_id: str = resp["result"]["id"]
    return snapshot_id


@snapshot_cli.command(name="create", help="Create a snapshot of the current project.")
@click.option("--description", help="A description of the snapshot", default=None)
@click.option(
    "--session-name",
    help="If specified, a snapshot of the remote session"
    "with that name will be taken.",
    default=None,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
@click.option(
    "--tag",
    type=str,
    help="Tag for this snapshot. Multiple tags can be specified by repeating this option.",
    multiple=True,
)
def snapshot_create(
    description: Optional[str], session_name: Optional[str], yes: bool, tag: List[str],
) -> None:
    project_definition = load_project_or_throw()
    project_id = get_project_id(project_definition.root)

    if session_name:
        # Create a remote snapshot.
        try:
            snapshot_id = remote_snapshot(
                project_id, session_name, tag, project_definition, description
            )
            print(
                "Snapshot {snapshot_id} of session {session_name} created!".format(
                    snapshot_id=snapshot_id, session_name=session_name
                )
            )
        except click.ClickException as e:
            raise e

    else:
        # Create a local snapshot.
        raise NotImplementedError("Local snapshotting is not supported anymore.")

    url = get_endpoint(f"/projects/{project_id}")
    print(f"Snapshot {snapshot_id} created. View at {url}")


@snapshot_cli.command(
    name="describe", help="Describe metadata and files of a snapshot."
)
@click.argument("name")
def snapshot_describe(name: str) -> None:
    try:
        description = describe_snapshot(name)
    except Exception as e:
        # Describing a snapshot can fail if the snapshot does not exist.
        raise click.ClickException(e)  # type: ignore

    print(description)



import copy
from typing import Any, Dict, Optional

import click


def format_command(command: str, parsed_args: Dict[Any, Any]) -> str:
    """Substitute arguments into command.
    Args:
        command (str): Shell comand with argument placeholders.
        parsed_args (dict): Dictionary that maps from argument names
            to their value.
    Returns:
        Shell command with parameters from parsed_args substituted.
    """
    for key, val in parsed_args.items():
        command = command.replace("{{" + key + "}}", str(val))
    return command


def get_session_runs(
    name: Optional[str], command: Optional[str], parsed_args: Dict[Any, Any]
) -> Any:
    """Get a list of sessions to start.
    Args:
        command (str): Shell command with argument placeholders.
        parsed_args (dict): Dictionary that maps from argument names
            to their values.
    Returns:
        List of sessions to start, which are dictionaries with keys:
            "name": Name of the session to start,
            "command": Command to run after starting the session,
            "params": Parameters for this run,
            "num_steps": 4 if a command should be run, 3 if not.
    """
    if not command:
        return [{"name": name, "command": None, "params": {}, "num_steps": 3}]

    # Try to find a wildcard argument (i.e. one that has a list of values)
    # and give an error if there is more than one (currently unsupported).
    wildcard_arg = None
    for key, val in parsed_args.items():
        if isinstance(val, list):
            if not wildcard_arg:
                wildcard_arg = key
            else:
                raise click.ClickException(
                    "More than one wildcard is not supported at the moment"
                )

    if not wildcard_arg:
        session_run = {
            "name": name,
            "command": format_command(command, parsed_args),
            "params": parsed_args,
            "num_steps": 4,
        }
        return [session_run]
    else:
        session_runs = []
        for val in parsed_args[wildcard_arg]:
            parsed_args = copy.deepcopy(parsed_args)
            parsed_args[wildcard_arg] = val
            session_run = {
                "name": "{}-{}-{}".format(name, wildcard_arg, val),
                "command": format_command(command, parsed_args),
                "params": parsed_args,
                "num_steps": 4,
            }
            session_runs.append(session_run)
        return session_runs



"""
Commands to interact with the Anyscale Projects API
"""

from typing import Optional

import click

from anyscale.api import get_anyscale_api_client
from anyscale.formatters import common_formatter
from anyscale.sdk.anyscale_client import CreateProject, UpdateProject  # type: ignore


@click.group(
    "projects", help="Commands to interact with the Projects API.",
)
def projects() -> None:
    pass


@projects.command(name="list", short_help="Lists all Projects the user has access to.")
@click.option(
    "--count", type=int, default=10, help="Number of projects to show. Defaults to 10."
)
@click.option(
    "--paging-token",
    required=False,
    help="Paging token used to fetch subsequent pages of projects.",
)
def list_projects(count: int, paging_token: Optional[str],) -> None:
    """Lists all non-deleted projects that the user has access to. """

    api_client = get_anyscale_api_client()
    response = api_client.list_projects(count=count, paging_token=paging_token)

    print(common_formatter.prettify_json(response.to_dict()))


@projects.command(name="get", short_help="Get details about a Project.")
@click.argument("project_id", required=True)
def get_project(project_id: str) -> None:
    """Get details about the Project with id PROJECT_ID"""

    api_client = get_anyscale_api_client()
    response = api_client.get_project(project_id)

    print(common_formatter.prettify_json(response.to_dict()))


@projects.command(name="create", short_help="Creates a Project.")
@click.argument("name", required=True)
@click.option(
    "--cluster_config",
    type=str,
    help="Cluster config associated with the Project. Default will be used if not provided.",
    required=False,
)
@click.option(
    "--description", type=str, help="Description of Project.", required=False,
)
def create_project(
    name: str, cluster_config: Optional[str], description: Optional[str]
) -> None:
    """Creates a Project with NAME, CLUSTER_CONFIG, and DESCRIPTION."""

    api_client = get_anyscale_api_client()
    create_data = CreateProject(
        name=name, cluster_config=cluster_config, description=description
    )
    response = api_client.create_project(create_data)

    print(common_formatter.prettify_json(response.to_dict()))


@projects.command(name="update", short_help="Updates a Project.")
@click.argument("project_id", required=True)
@click.option(
    "--cluster_config",
    type=str,
    help="Cluster config associated with the Project. No changes will be made if not provided.",
    required=False,
)
@click.option(
    "--description",
    type=str,
    help="Description of Project. No changes will be made if not provided.",
    required=False,
)
def update_project(
    project_id: str, cluster_config: Optional[str], description: Optional[str]
) -> None:
    """
    Updates Project PROJECT_ID with CLUSTER_CONFIG and DESCRIPTION if provided. No changes will be made if
    a field is not provided.
    """

    api_client = get_anyscale_api_client()
    update_data = UpdateProject(cluster_config=cluster_config, description=description)
    response = api_client.update_project(project_id, update_data)

    print(common_formatter.prettify_json(response.to_dict()))


@projects.command(name="delete", short_help="Deletes a Project.")
@click.argument("project_id", required=True)
def delete_project(project_id: str) -> None:
    """Delete the Project with id PROJECT_ID"""

    api_client = get_anyscale_api_client()
    api_client.delete_project(project_id)



import datetime
import logging
from typing import Dict, List, Optional

import click

from anyscale.auth_proxy import app as auth_proxy_app
import anyscale.legacy_projects as ray_scripts
from anyscale.project import (
    get_project_id,
    get_project_session,
    load_project_or_throw,
)
from anyscale.snapshot import (
    copy_file,
    get_snapshot_id,
)
from anyscale.util import (
    deserialize_datetime,
    execution_log_name,
    get_endpoint,
    send_json_request,
)


logger = logging.getLogger(__file__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)


@click.group("session", help="Commands for working with sessions.", hidden=True)
def session_cli() -> None:
    pass


@session_cli.command(name="logs", help="Show logs for the current session.")
@click.option("--name", help="Name of the session to run this command on", default=None)
@click.option("--command-id", help="ID of the command to get logs for", default=None)
def session_logs(name: Optional[str], command_id: Optional[int]) -> None:
    project_definition = load_project_or_throw()
    project_id = get_project_id(project_definition.root)
    # If the command_id is not specified, determine it by getting the
    # last run command from the active session.
    if not command_id:
        session = get_project_session(project_id, name)
        resp = send_json_request(
            "/api/v2/session_commands/?session_id={}".format(session["id"]), {}
        )
        # Search for latest run command
        last_created_at = datetime.datetime.min
        last_created_at = last_created_at.replace(tzinfo=datetime.timezone.utc)
        for command in resp["results"]:
            created_at = deserialize_datetime(command["created_at"])
            if created_at > last_created_at:
                last_created_at = created_at
                command_id = command["id"]
        if not command_id:
            raise click.ClickException(
                "No comand was run yet on the latest active session {}".format(
                    session["name"]
                )
            )
    resp_out = send_json_request(
        "/api/v2/session_commands/{session_command_id}/execution_logs".format(
            session_command_id=command_id
        ),
        {"log_type": "out", "start_line": 0, "end_line": 1000000000},
    )
    resp_err = send_json_request(
        "/api/v2/session_commands/{session_command_id}/execution_logs".format(
            session_command_id=command_id
        ),
        {"log_type": "err", "start_line": 0, "end_line": 1000000000},
    )
    # TODO(pcm): We should have more options here in the future
    # (e.g. show only stdout or stderr, show only the tail, etc).
    print("stdout:")
    print(resp_out["result"]["lines"])
    print("stderr:")
    print(resp_err["result"]["lines"])


# Note: These functions are not normally updated, please cc yiran or ijrsvt on changes.
@session_cli.command(
    name="upload_command_logs", help="Upload logs for a command.", hidden=True
)
@click.option(
    "--command-id", help="ID of the command to upload logs for", type=str, default=None
)
def session_upload_command_logs(command_id: Optional[str]) -> None:
    resp = send_json_request(
        "/api/v2/session_commands/{session_command_id}/upload_logs".format(
            session_command_id=command_id
        ),
        {},
        method="POST",
    )
    assert resp["result"]["session_command_id"] == command_id

    allowed_sources = [
        execution_log_name(command_id) + ".out",
        execution_log_name(command_id) + ".err",
    ]

    for source, target in resp["result"]["locations"].items():
        if source in allowed_sources:
            copy_file(True, source, target, download=False)


# Note: These functions are not normally updated, please cc yiran or ijrsvt on changes.
@session_cli.command(
    name="finish_command", help="Finish executing a command.", hidden=True
)
@click.option(
    "--command-id", help="ID of the command to finish", type=str, required=True
)
@click.option(
    "--stop", help="Stop session after command finishes executing.", is_flag=True
)
@click.option(
    "--terminate",
    help="Terminate session after command finishes executing.",
    is_flag=True,
)
def session_finish_command(command_id: str, stop: bool, terminate: bool) -> None:
    with open(execution_log_name(command_id) + ".status") as f:
        status_code = int(f.read().strip())
    send_json_request(
        f"/api/v2/session_commands/{command_id}/finish",
        {"status_code": status_code, "stop": stop, "terminate": terminate},
        method="POST",
    )


# Note: These functions are not normally updated, please cc yiran or ijrsvt on changes.
@session_cli.command(name="auth_start", help="Start the auth proxy", hidden=True)
def auth_start() -> None:
    from aiohttp import web

    web.run_app(auth_proxy_app)


@session_cli.command(
    name="web_terminal_server", help="Start the web terminal server", hidden=True
)
@click.option(
    "--deploy-environment",
    help="Anyscale deployment type (development, test, staging, production)",
    type=str,
    required=True,
)
@click.option(
    "--use-debugger", help="Activate the Anyscale debugger.", is_flag=True,
)
def web_terminal_server(deploy_environment: str, use_debugger: bool) -> None:
    from anyscale.webterminal.webterminal import main

    main(deploy_environment, use_debugger)


@click.command(
    name="start",
    context_settings=dict(ignore_unknown_options=True,),
    help="Start a session based on the current project configuration.",
    hidden=True,
)
@click.option("--session-name", help="The name of the created session.", default=None)
# TODO(pcm): Change this to be
# anyscale session start --arg1=1 --arg2=2 command args
# instead of
# anyscale session start --session-args=--arg1=1,--arg2=2 command args
@click.option(
    "--session-args",
    help="Arguments that get substituted into the cluster config "
    "in the format --arg1=1,--arg2=2",
    default="",
)
@click.option(
    "--snapshot",
    help="If set, start the session from the given snapshot.",
    default=None,
)
@click.option(
    "--config",
    help="If set, use this cluster file rather than the default"
    " listed in project.yaml.",
    default=None,
)
@click.option(
    "--min-workers",
    help="Overwrite the minimum number of workers in the cluster config.",
    default=None,
)
@click.option(
    "--max-workers",
    help="Overwrite the maximum number of workers in the cluster config.",
    default=None,
)
@click.option(
    "--run", help="Command to run.", default=None,
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.option("--cloud-id", help="Id of the cloud to use", default=None)
@click.option("--cloud-name", help="Name of the cloud to use", default=None)
def anyscale_start(
    session_args: str,
    snapshot: Optional[str],
    session_name: Optional[str],
    config: Optional[str],
    min_workers: Optional[int],
    max_workers: Optional[int],
    run: Optional[str],
    args: List[str],
    cloud_id: Optional[str],
    cloud_name: Optional[str],
) -> None:
    command_name = run

    project_definition = load_project_or_throw()
    project_id = get_project_id(project_definition.root)

    if cloud_id and cloud_name:
        raise click.ClickException("Please provide either cloud id or cloud name.")
    elif cloud_name:
        resp_get_cloud = send_json_request(
            "/api/v2/clouds/find_by_name", {"name": cloud_name}, method="POST"
        )
        cloud = resp_get_cloud["result"]
        cloud_id = cloud["id"]

    if not session_name:
        session_list = send_json_request(
            "/api/v2/sessions/", {"project_id": project_id, "active_only": False}
        )["results"]
        session_name = "session-{0}".format(len(session_list) + 1)

    # Parse the session arguments.
    if config:
        project_definition.config["cluster"]["config"] = config

    session_params: Dict[str, str] = {}

    if command_name:
        command_name = " ".join([command_name] + list(args))
    session_runs = ray_scripts.get_session_runs(session_name, command_name, {})

    assert len(session_runs) == 1, "Running sessions with a wildcard is deprecated"
    session_run = session_runs[0]

    snapshot_id = None
    if snapshot is not None:
        snapshot_id = get_snapshot_id(project_definition.root, snapshot)

    session_name = session_run["name"]
    resp = send_json_request(
        "/api/v2/sessions/",
        {"project_id": project_id, "name": session_name, "active_only": False},
    )
    if len(resp["results"]) == 0:
        resp = send_json_request(
            "/api/v2/sessions/create_new_session",
            {
                "project_id": project_id,
                "name": session_name,
                "snapshot_id": snapshot_id,
                "session_params": session_params,
                "command_name": command_name,
                "command_params": session_run["params"],
                "shell": True,
                "min_workers": min_workers,
                "max_workers": max_workers,
                "cloud_id": cloud_id,
            },
            method="POST",
        )
    elif len(resp["results"]) == 1:
        if session_params != {}:
            raise click.ClickException(
                "Session parameters are not supported when restarting a session"
            )
        send_json_request(
            "/api/v2/sessions/{session_id}/start".format(
                session_id=resp["results"][0]["id"]
            ),
            {"min_workers": min_workers, "max_workers": max_workers},
            method="POST",
        )
    else:
        raise click.ClickException(
            "Multiple sessions with name {} exist".format(session_name)
        )
    # Print success message
    url = get_endpoint(f"/projects/{project_id}")
    print(f"Session {session_name} starting. View progress at {url}")



import inspect
import os
from sys import path
from typing import Any, Dict, List

anyscale_dir = os.path.dirname(os.path.abspath(__file__))
path.append(os.path.join(anyscale_dir, "client"))
path.append(os.path.join(anyscale_dir, "sdk"))
ANYSCALE_RAY_DIR = os.path.join(anyscale_dir, "anyscale_ray")

from anyscale.connect import SessionBuilder  # noqa: E402

# Auto-add all Anyscale connect builder functions to the top-level.
for attr, _ in inspect.getmembers(SessionBuilder, inspect.isfunction):
    if attr.startswith("_"):
        continue

    def _new_builder(attr: str) -> Any:
        def new_session_builder(*a: List[Any], **kw: Dict[str, Any]) -> Any:
            """Start building an Anyscale session.

            See ``anyscale.SessionBuilder`` for documentation of this
            experimental feature.
            """

            builder = SessionBuilder()
            return getattr(SessionBuilder, attr)(builder, *a, **kw)

        return new_session_builder

    globals()[attr] = _new_builder(attr)

__version__ = "0.3.34"

ANYSCALE_ENV = os.environ.copy()
ANYSCALE_ENV["PYTHONPATH"] = ANYSCALE_RAY_DIR + ":" + ANYSCALE_ENV.get("PYTHONPATH", "")



import json
import os
from typing import Dict

import click

import anyscale


def load_credentials() -> str:
    # The environment variable ANYSCALE_CLI_TOKEN can be used to
    # overwrite the credentials in ~/.anyscale/credentials.json
    if "ANYSCALE_CLI_TOKEN" in os.environ:
        return os.environ["ANYSCALE_CLI_TOKEN"]
    path = os.path.expanduser("~/.anyscale/credentials.json")
    if not os.path.exists(path):
        raise click.ClickException(
            "Credentials not found. You need to create an account at {0} "
            "and then go to {0}/credentials and follow the instructions.".format(
                anyscale.conf.ANYSCALE_HOST
            )
        )
    with open(path) as f:
        try:
            credentials: Dict[str, str] = json.load(f)
        except json.JSONDecodeError:
            msg = (
                "Unable to load user credentials.\n\nTip: Try creating your "
                "user credentials again by going to {}/credentials and "
                "following the instructions. If this does not work, "
                "please contact Anyscale support!".format(anyscale.conf.ANYSCALE_HOST)
            )
            raise click.ClickException(msg)
    if "cli_token" not in credentials:
        raise click.ClickException("Credential file not valid, please regenerate it.")
    return credentials["cli_token"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# mypy: ignore-errors

import os
import re
import shutil
import subprocess
import sys
from typing import List
import zipfile

from setuptools import Distribution, find_packages, setup
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.sdist import sdist as sdist_orig


def find_version(path):
    with open(path) as f:
        match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.MULTILINE,
        )
        if match:
            return match.group(1)
        raise RuntimeError("Unable to find version string.")


def move_file(filename):
    # TODO(rkn): This feels very brittle. It may not handle all cases. See
    # https://github.com/apache/arrow/blob/master/python/setup.py for an
    # example.
    source = filename

    destination = os.path.join(os.path.dirname(__file__), filename)
    # Create the target directory if it doesn't already exist.
    parent_directory = os.path.dirname(destination)
    if not os.path.exists(parent_directory):
        os.makedirs(parent_directory)
    if not os.path.exists(destination):
        print("Copying {} to {}.".format(source, destination))
        shutil.copy(source, destination, follow_symlinks=True)


def download_and_copy_files(url: str, files_to_copy: List[str]) -> None:
    """
    Downloads an archive from S3, unzips it and then copies the file to the anyscale folder.
    """
    import io
    import requests
    import tempfile

    work_dir = tempfile.mkdtemp()
    try:
        content = requests.get(url).content
        archive = zipfile.ZipFile(io.BytesIO(content))
        archive.extractall(pwd=work_dir.encode())
        for f in files_to_copy:
            destination = os.path.join("anyscale", f)
            # Remove the file if it already exists to make sure old
            # versions get removed.
            try:
                os.remove(destination)
            except OSError:
                pass
            shutil.copy2(f, destination)
            os.chmod(destination, 0o755)
            move_file(destination)
    finally:
        shutil.rmtree(work_dir)


def download_and_copy_fswatch():
    download_and_copy_files(
        "https://anyscale-dev.s3-us-west-2.amazonaws.com/fswatch-1.14.0-2.zip",
        ["fswatch-linux", "fswatch-darwin", "libfswatch.11.dylib"],
    )


class SdistCommand(sdist_orig):
    def run(self):
        download_and_copy_fswatch()
        # run original sdist
        super().run()


class BinaryDistribution(Distribution):
    def is_pure(self):
        return True

    def has_ext_modules(self):
        return True


RAY_COMMIT = "aba9288615a5769541a082689a1d08e9e08c553f"
RAY_VERSION = "ray-1.1.0.dev0"


RAY_WHEELS = {
    "linux": {
        "3.8": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp38-cp38-manylinux2014_x86_64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
        "3.7": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp37-cp37m-manylinux2014_x86_64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
        "3.6": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp36-cp36m-manylinux2014_x86_64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
    },
    "darwin": {
        "3.8": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp38-cp38-macosx_10_13_x86_64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
        "3.7": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp37-cp37m-macosx_10_13_intel.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
        "3.6": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp36-cp36m-macosx_10_13_intel.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
    },
    "win32": {
        "3.8": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp38-cp38-win_amd64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
        "3.7": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp37-cp37m-win_amd64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
        "3.6": "https://s3-us-west-2.amazonaws.com/ray-wheels/master/{}/{}-cp36-cp36m-win_amd64.whl".format(
            RAY_COMMIT, RAY_VERSION
        ),
    },
}


def install_private_ray(anyscale_path):
    platform = sys.platform
    py_version = "{0}.{1}".format(*sys.version_info[:2])

    matching_wheel = None
    for target_platform, wheel_map in RAY_WHEELS.items():
        print(f"Evaluating os={target_platform}, python={list(wheel_map)}")
        if platform.startswith(target_platform):
            if py_version in wheel_map:
                matching_wheel = wheel_map[py_version]
                break
        print("Not matched.")

    if matching_wheel is None:
        raise Exception(
            "Unable to identify a matching Ray wheel for Platform: {}, Python: {}".format(
                platform, py_version
            )
        )

    cmd = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--target={}".format(os.path.join(anyscale_path, "anyscale", "anyscale_ray")),
        "-U",
        matching_wheel,
    ]
    print(f"Running: {' '.join(cmd)}.")
    subprocess.check_call(cmd)  # noqa: B1


def install_private_official_release_ray(anyscale_path):
    cmd = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--target={}".format(os.path.join(anyscale_path, "anyscale", "anyscale_ray")),
        "-U",
        "ray==1.1.0",
    ]
    print(f"Running: {' '.join(cmd)}.")
    subprocess.check_call(cmd)  # noqa: B1


class PostDevelopCommand(develop):
    def run(self):
        develop.run(self)
        print("Installing private version of Ray")
        download_and_copy_fswatch()
        install_private_official_release_ray(os.path.dirname(__file__))


class PostInstallCommand(install):
    def run(self):
        install.run(self)
        print("Installing private version of Ray")
        install_private_official_release_ray(self.install_lib)


extras_require = {"backend": ["terminado", "tornado"]}


setup(
    name="anyscale",
    version=find_version("anyscale/__init__.py"),
    author="Anyscale Inc.",
    description=("Command Line Interface for Anyscale"),
    packages=find_packages(exclude="tests"),
    cmdclass={
        "sdist": SdistCommand,
        "develop": PostDevelopCommand,
        "install": PostInstallCommand,
    },
    distclass=BinaryDistribution,
    setup_requires=["setuptools_scm"],
    data_files=[
        (
            "generated",
            [
                "anyscale/fswatch-linux",
                "anyscale/fswatch-darwin",
                "anyscale/libfswatch.11.dylib",
            ],
        ),
    ],
    install_requires=[
        "aiohttp",
        "aiohttp_middlewares",
        "boto3",
        "Click>=7.0",
        "expiringdict",
        "GitPython",
        "jinja2",
        "jsonpatch",
        "jsonschema",
        "packaging",
        "ray>=1.0.1",
        "requests",
        "tabulate",
        "wrapt",
    ],
    extras_require=extras_require,
    entry_points={"console_scripts": ["anyscale=anyscale.scripts:main"]},
    include_package_data=True,
    zip_safe=False,
)



"""
Defines and implements `anyscale list {resource}` commands.
Supported resources: clouds, projects, sessions, ips
"""

from typing import Any, Optional

import click

from anyscale.controllers.list_controller import ListController


@click.group("list", help="List resources (projects, sessions) within Anyscale.")
def list_cli() -> None:
    pass


@list_cli.command(
    name="clouds", help="List the clouds currently available in your account."
)
@click.option("--json", "show_json", help="Return the results in json", is_flag=True)
def list_clouds(show_json: bool) -> None:
    list_controller = ListController()
    output = list_controller.list_clouds(json_format=show_json)
    print(output)


@list_cli.command(name="projects", help="List all accessible projects.")
@click.option("--json", "show_json", help="Return the results in json", is_flag=True)
@click.pass_context
def project_list(ctx: Any, show_json: bool) -> None:
    list_controller = ListController()
    output = list_controller.list_projects(json_format=show_json)
    print(output)


@list_cli.command(name="sessions", help="List all sessions within the current project.")
@click.option(
    "--name",
    help="Name of the session. If provided, this prints the snapshots that "
    "were applied and commands that ran for all sessions that match "
    "this name.",
    default=None,
)
@click.option("--all", help="List all sessions, including inactive ones.", is_flag=True)
@click.option("--json", "show_json", help="Return the results in json", is_flag=True)
def session_list(name: Optional[str], all: bool, show_json: bool) -> None:
    list_controller = ListController()
    output = list_controller.list_sessions(
        name=name, show_all=all, json_format=show_json
    )
    print(output)


@list_cli.command(name="ips", help="List IP addresses of head and worker nodes.")
@click.argument("session-name", required=False, type=str)
@click.option("--json", "show_json", help="Return the results in json", is_flag=True)
@click.option(
    "--all", "all_sessions", help="List IPs of all active sessions.", is_flag=True
)
def list_ips(session_name: Optional[str], show_json: bool, all_sessions: bool) -> None:
    """List IP addresses of head and worker nodes."""
    list_controller = ListController()
    output = list_controller.list_ips(
        session_name=session_name, json_format=show_json, all_sessions=all_sessions
    )
    print(output)
